#!/usr/bin/env python3
#
# Tests the Model class.
#
# Notes:
#  - Tests for dependency checking in models are in `test_dependency_checking`.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import pickle
import re
import unittest

import myokit

from myokit.tests import TemporaryDirectory, WarningCollector


class ModelTest(unittest.TestCase):
    """
    Tests parts of :class:`myokit.Model`.
    """

    def test_add_component_allow_renamining(self):
        # Test the ``Model.add_component_allow_renaming`` method.

        m = myokit.Model('test')
        c = m.add_component('c')
        self.assertTrue(m.has_component('c'))
        self.assertRaises(myokit.DuplicateName, m.add_component, 'c')
        d = m.add_component_allow_renaming('c')
        self.assertEqual(c.name(), 'c')
        self.assertEqual(d.name(), 'c_1')
        e = m.add_component_allow_renaming('c')
        self.assertEqual(e.name(), 'c_2')

        # Test repeated calls
        r = m.add_component('r')
        for i in range(10):
            r = m.add_component_allow_renaming('r')
            self.assertEqual(r.name(), 'r_' + str(1 + i))

    def test_add_function(self):
        # Test the ``Model.add_function`` method.

        m = myokit.Model('m')
        c = m.add_component('c')
        x = c.add_variable('x')

        # Test basics
        m.add_function('f', ('a', 'b', 'c'), 'a + b + c')
        x.set_rhs('f(1, 2, 3)')
        self.assertEqual(x.eval(), 6)

        # Test duplicate name
        # Different number of arguments is allowed:
        m.add_function('f', ('a', 'b'), 'a + b')
        self.assertRaisesRegex(
            myokit.DuplicateFunctionName, 'already defined', m.add_function,
            'f', ('a', 'b'), 'a - b')

        # Test duplicate argument name
        self.assertRaisesRegex(
            myokit.DuplicateFunctionArgument, 'already in use',
            m.add_function, 'g', ('a', 'a'), 'a + a')

        # Dot operator is not allowed, nor are init or partial
        self.assertRaisesRegex(
            myokit.InvalidFunction, r'dot\(\) operator',
            m.add_function, 'fdot', ('a', ), 'dot(a)')
        a = myokit.Name('a')
        self.assertRaisesRegex(
            myokit.InvalidFunction, r'partial\(\) operator',
            m.add_function, 'fpart', ('a', ), myokit.PartialDerivative(a, a))
        self.assertRaisesRegex(
            myokit.InvalidFunction, r'init\(\) operator',
            m.add_function, 'finit', ('a', ), myokit.InitialValue(a))

        # Unused argument
        self.assertRaisesRegex(
            myokit.InvalidFunction, 'never used', m.add_function, 'fun',
            ('a', 'b'), 'a')

        # Unspecified variable
        self.assertRaisesRegex(
            myokit.InvalidFunction, 'never declared',
            m.add_function, 'fun', ('a', ), 'a + b')

    def test_binding(self):
        # Tests setting and getting of bindings

        # Note that set_binding() is part of Variable, so not tested here
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_binding('hello')
        y = c.add_variable('y')
        y.set_binding('goodbye')
        z = c.add_variable('z')
        z.set_binding('x')

        # Test binding()
        self.assertEqual(m.binding('goodbye'), y)
        self.assertEqual(m.binding('hello'), x)
        self.assertEqual(m.binding('x'), z)
        self.assertEqual(m.binding('y'), None)

        # Test bindings()
        bindings = dict(m.bindings())
        self.assertEqual(bindings['goodbye'], y)
        self.assertEqual(bindings['hello'], x)
        self.assertEqual(bindings['x'], z)
        self.assertFalse('y' in bindings)

        # Test bindingx()
        self.assertEqual(m.bindingx('goodbye'), y)
        self.assertEqual(m.bindingx('hello'), x)
        self.assertEqual(m.bindingx('x'), z)
        self.assertRaisesRegex(
            myokit.IncompatibleModelError,
            'No variable found with binding "y"',
            m.bindingx, 'y')

    def test_bindings(self):
        # Test setting bindings and :meth:`Model.bindings()`.

        # Test set_binding() and bindings()
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        v = c.add_variable('v')
        v.set_rhs('3 - v')
        w = c.add_variable('w')
        w.set_rhs(0)
        bindings = list(m.bindings())
        self.assertEqual(len(bindings), 1)
        self.assertEqual(bindings[0][0], 'time')
        self.assertEqual(bindings[0][1], t)

        # Can't have two labels
        self.assertRaisesRegex(
            myokit.InvalidBindingError, 'already bound to', t.set_binding,
            'bert')

        # No two variables can have the same label
        self.assertRaisesRegex(
            myokit.InvalidBindingError, 'Duplicate binding', v.set_binding,
            'time')

        # Binding can't overlap with label
        v.set_label('membrane_potential')
        self.assertRaisesRegex(
            myokit.InvalidBindingError, 'in use as a label', w.set_binding,
            'membrane_potential')

        # State variables can't be bound
        v.promote(0)
        self.assertRaisesRegex(
            myokit.InvalidBindingError, 'State variables', v.set_binding, 'x')

    def test_check_units(self):
        # Test the ``model.check_units`` method.

        model = myokit.Model('m')
        component = model.add_component('c')
        t = component.add_variable('time')
        t.set_binding('time')

        # Check units before any rhs or units set
        s = myokit.UNIT_STRICT
        self.assertRaisesRegex(
            myokit.IntegrityError, 'No RHS set', model.check_units)
        self.assertRaisesRegex(
            myokit.IntegrityError, 'No RHS set', model.check_units, s)

        # Check units before any rhs set
        t.set_unit('s')
        self.assertRaisesRegex(
            myokit.IntegrityError, 'No RHS set', model.check_units)
        self.assertRaisesRegex(
            myokit.IntegrityError, 'No RHS set', model.check_units, s)

        # Check mini model with rhs and units, no states
        t.set_rhs('0 [s]')
        a = component.add_variable('a')
        a.set_rhs(1)
        model.check_units()
        model.check_units(s)

        # Check mini model with a state
        # Strict check should fail: a's RHS should be in 1/s
        a.promote(0)
        model.check_units()
        self.assertRaises(myokit.IncompatibleUnitError, model.check_units, s)
        a.set_rhs('1 [1/s]')
        model.check_units(s)

        b = component.add_variable('b')
        b.set_rhs(2)
        c = component.add_variable('c')
        c.set_rhs('2 * b')
        model.check_units()
        model.check_units(s)

        a.set_rhs('1 [N/s]')
        b.set_rhs('2 [m]')
        c.set_rhs('a * b')

        # No variable units set
        model.check_units()
        self.assertRaises(myokit.IncompatibleUnitError, model.check_units, s)

        # Variable unit set for state
        a.set_unit('N')     # So rhs should be N/s
        b.set_unit('m')
        model.check_units()
        self.assertRaises(myokit.IncompatibleUnitError, model.check_units, s)

        # Bad derived unit
        c.set_unit('A')
        self.assertRaises(myokit.IncompatibleUnitError, model.check_units)
        self.assertRaises(myokit.IncompatibleUnitError, model.check_units, s)

        c.set_unit(myokit.parse_unit('N*m'))
        model.check_units()
        model.check_units(s)

        # References use variable unit, not RHS unit!
        model = myokit.Model('m')
        component = model.add_component('c')
        x = component.add_variable('x')
        y = component.add_variable('y')
        x.set_unit(None)
        y.set_unit(None)
        x.set_rhs('5 [mV]')
        y.set_rhs('3 [A] + x')  # x unit is unspecified, not mV!
        model.check_units()
        self.assertRaises(myokit.IncompatibleUnitError, model.check_units, s)

        # Tokens are used in IncompatibleUnitError messages
        m = myokit.parse_model('\n'.join([
            '[[model]]',
            '[a]',
            't = 0 [ms] bind time',
        ]))
        try:
            m.check_units(s)
        except myokit.IncompatibleUnitError as e:
            self.assertIn('on line 3', str(e))
            token = e.token()
            self.assertIsNotNone(token)
            self.assertEqual(token[2], 3)
            self.assertEqual(token[3], 0)

        # Test comparison with floating point issues
        m = myokit.parse_model('\n'.join([
            '[[model]]',
            '[a]',
            'x = 1 [cm^3] bind time',
            '    in [cm^3]',
            'y = 2 [day]',
            '    in [day]',
            'z = 3 [day^3]',
            '    in [day^3]',
            'a = (x / y / y / y) * z',
            '    in [cm^3]',
        ]))
        m.check_units(s)

    def test_clone(self):
        # Test :meth:`Model.clone() and :meth:`Model.has_parse_info()`.

        # Test model, component, variables
        m1 = myokit.load_model('example')
        m2 = m1.clone()
        self.assertFalse(m1 is m2)
        self.assertNotEqual(m1, m2)
        self.assertTrue(m1.is_similar(m2, True))

        # Test unames and uname prefixes
        m1.reserve_unique_names('barnard', 'lincoln', 'glasgow')
        m1.reserve_unique_name_prefix('monkey', 'giraffe')
        m1.reserve_unique_name_prefix('ostrich', 'turkey')
        m2 = m1.clone()
        self.assertTrue(m1.is_similar(m2, True))

        # Test tokens are not cloned
        self.assertTrue(m1.has_parse_info())
        self.assertFalse(m2.has_parse_info())

        # Test initial value expressions are maintained
        m1 = myokit.Model('moo')
        component = m1.add_component('c')
        p = component.add_variable('p')
        q = component.add_variable('q')
        r = component.add_variable('r')
        s = component.add_variable('s')
        z = component.add_variable('z')
        z.set_rhs(123)
        p.promote(1)
        q.promote('1 + exp(3)')
        r.promote('c.z')
        s.promote('1 [g] + c.z / 2 [1/g]')
        m2 = m1.clone()
        self.assertEqual(m2.get('c.p').initial_value(), myokit.Number(1))
        self.assertEqual(m2.get('c.q').initial_value().code(), '1 + exp(3)')
        self.assertEqual(m2.get('c.r').initial_value(), m2.get('c.z').lhs())
        self.assertEqual(m2.get('c.s').initial_value().code(),
                         '1 [g] + c.z / 2 [1/g]')

    def test_code(self):
        # Test :meth:`Model.code()`.

        model = myokit.Model('m')
        component = model.add_component('comp1')
        a = component.add_variable('a')
        b = component.add_variable('b')
        c = component.add_variable('c')
        a.set_rhs('1 [N]')
        a.set_label('aaa')
        b.set_rhs('2 [m]')
        b.set_binding('bbb')
        c.set_rhs('a * b')
        c.set_unit('N*m')
        component2 = model.add_component('comp2')
        d = component2.add_variable('d')
        d.set_rhs(myokit.Name(a))

        self.assertEqual(
            model.code(),
            '[[model]]\n'
            'name: m\n'
            '\n'
            '[comp1]\n'
            'a = 1 [N] label aaa\n'
            'b = 2 [m] bind bbb\n'
            'c = a * b\n'
            '    in [J]\n'
            '\n'
            '[comp2]\n'
            'd = comp1.a\n'
            '\n'
        )

        a.set_rhs('1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + 11')
        b.set_rhs('2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + 11 + 12')

        self.assertEqual(
            model.code(line_numbers=True),
            ' 1 [[model]]\n'
            ' 2 name: m\n'
            ' 3 \n'
            ' 4 [comp1]\n'
            ' 5 a = 1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + 11\n'
            ' 6     label aaa\n'
            ' 7 b = 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + 11 + 12\n'
            ' 8     bind bbb\n'
            ' 9 c = a * b\n'
            '10     in [J]\n'
            '11 \n'
            '12 [comp2]\n'
            '13 d = comp1.a\n'
        )

        # Test initial value expressions are maintained
        model = myokit.Model('moo')
        component = model.add_component('c')
        p = component.add_variable('p')
        q = component.add_variable('q')
        r = component.add_variable('r')
        s = component.add_variable('s')
        z = component.add_variable('z')
        z.set_rhs(123)
        p.promote(1)
        q.promote('1 + exp(3)')
        r.promote('c.z')
        s.promote('1 [g] + c.z / 2 [1/g]')

        p.set_rhs(1)
        q.set_rhs(2)
        r.set_rhs(3)
        s.set_rhs('4 [g/s]')

        self.assertEqual(model.code(line_numbers=False), '\n'.join([
            '[[model]]',
            'name: moo',
            '# Initial values',
            'c.p = 1',
            'c.q = 1 + exp(3)',
            'c.r = c.z',
            'c.s = 1 [g] + c.z / 2 [1/g]',
            '',
            '[c]',
            'dot(p) = 1',
            'dot(q) = 2',
            'dot(r) = 3',
            'dot(s) = 4 [g/s]',
            'z = 123\n\n',
        ]))

    def test_evaluate_derivatives(self):
        # Test Model.evaluate_derivatives().
        model = myokit.Model('m')
        component = model.add_component('comp1')
        t = component.add_variable('time')
        t.set_binding('time')
        t.set_rhs(1)
        a = component.add_variable('a')
        b = component.add_variable('b')
        c = component.add_variable('c')
        d = component.add_variable('d')
        a.promote(1)
        a.set_rhs('1')
        b.promote(2)
        b.set_rhs('2 * b')
        c.promote(3)
        c.set_rhs('b + d')
        d.set_rhs('1 * c')
        model.validate()

        # Test without input
        self.assertEqual(model.evaluate_derivatives(), [1, 4, 5])
        self.assertEqual(model.evaluate_derivatives(ignore_errors=True),
                         [1, 4, 5])

        # Test with alternative state
        self.assertEqual(
            model.evaluate_derivatives(state=[2, 3, 4]), [1, 6, 7])
        self.assertEqual(
            model.evaluate_derivatives(state=[2, 3, 4], ignore_errors=True),
            [1, 6, 7])

        # Test with input
        c.set_rhs('b + c + time')
        self.assertEqual(model.evaluate_derivatives(), [1, 4, 6])
        self.assertEqual(
            model.evaluate_derivatives(ignore_errors=True), [1, 4, 6])
        self.assertEqual(
            model.evaluate_derivatives(state=[2, 3, 4]), [1, 6, 8])
        self.assertEqual(
            model.evaluate_derivatives(state=[2, 3, 4], ignore_errors=True),
            [1, 6, 8])
        self.assertEqual(
            model.evaluate_derivatives(state=[2, 3, 4], inputs={'time': 10}),
            [1, 6, 17])
        self.assertEqual(
            model.evaluate_derivatives(
                state=[2, 3, 4], inputs={'time': 10}, ignore_errors=True),
            [1, 6, 17])

        # Deprecated name
        with WarningCollector() as w:
            self.assertEqual(
                model.eval_state_derivatives(
                    state=[1, 1, 2], inputs={'time': 0}),
                [1, 2, 3])
        self.assertIn('deprecated', w.text())

        # Ignoring errors should lead to Nans
        c.set_rhs('(b + c) / 0')
        self.assertRaises(myokit.NumericalError, model.evaluate_derivatives)
        nan = model.evaluate_derivatives(ignore_errors=True)[2]
        self.assertNotEqual(nan, nan)   # x != x is a nan test...

    def test_format_state(self):
        # Test Model.format_state()

        self.maxDiff = None
        m = myokit.load_model('example')

        # Test without state argument
        self.assertEqual(
            m.format_state(),
            'membrane.V = -84.5286\n'
            'ina.m      = 0.0017\n'
            'ina.h      = 0.9832\n'
            'ina.j      = 0.995484\n'
            'ica.d      = 3e-06\n'
            'ica.f      = 1.0\n'
            'ik.x       = 0.0057\n'
            'ica.Ca_i   = 0.0002'
        )

        # Test with state argument
        state1 = [1, 2, 3, 4, 5, 6, 7, 8]
        state1[3] = 124.35624574537437
        self.assertEqual(
            m.format_state(state1),
            'membrane.V = 1\n'
            'ina.m      = 2\n'
            'ina.h      = 3\n'
            'ina.j      =  1.24356245745374366e+02\n'
            'ica.d      = 5\n'
            'ica.f      = 6\n'
            'ik.x       = 7\n'
            'ica.Ca_i   = 8'
        )

        # Test with invalid state argument
        self.assertRaisesRegex(
            ValueError, r'sequence of \(8\)', m.format_state, [1, 2, 3])

        # Test with precision argument
        state1 = [1, 2, 3, 4, 5, 6, 7, 8]
        state1[3] = 124.35624574537437
        self.assertEqual(
            m.format_state(state1, precision=myokit.SINGLE_PRECISION),
            'membrane.V = 1\n'
            'ina.m      = 2\n'
            'ina.h      = 3\n'
            'ina.j      =  1.243562457e+02\n'
            'ica.d      = 5\n'
            'ica.f      = 6\n'
            'ik.x       = 7\n'
            'ica.Ca_i   = 8'
        )

        # Test with second state argument
        self.assertEqual(
            m.format_state([1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1]),
            'membrane.V = 1    8\n'
            'ina.m      = 2    7\n'
            'ina.h      = 3    6\n'
            'ina.j      = 4    5\n'
            'ica.d      = 5    4\n'
            'ica.f      = 6    3\n'
            'ik.x       = 7    2\n'
            'ica.Ca_i   = 8    1'
        )

        # Test with invalid second state argument
        self.assertRaisesRegex(
            ValueError, r'sequence of \(8\)', m.format_state,
            [1, 2, 3, 4, 5, 6, 7, 8], [1, 2, 3])

        # Test with model that has expressions in its initial state.
        # These will be evaluated (the method handles states, not initial state
        # expressions).
        m.get('ina.m').set_initial_value('sqrt(4) / 4')
        m.get('ina.h').set_initial_value('ina.gNa / 20')
        self.assertEqual(
            m.format_state(),
            'membrane.V = -84.5286\n'
            'ina.m      = 0.5\n'
            'ina.h      = 0.8\n'
            'ina.j      = 0.995484\n'
            'ica.d      = 3e-06\n'
            'ica.f      = 1.0\n'
            'ik.x       = 0.0057\n'
            'ica.Ca_i   = 0.0002'
        )

    def test_format_state_derivatives(self):
        # Test Model.format_state_derivatives().

        self.maxDiff = None
        m = myokit.load_model('example')

        # For whatever reason, CI gives slightly different final digits some
        # times.
        r1 = re.compile(
            r'^[a-zA-Z]\w*\.[a-zA-Z]\w*\s+=\s+([^\s]+)\s+dot\s+=\s+([^\s]+)$')

        def almost_equal(x, y):
            if x == y:
                return True
            g1, g2 = r1.match(x), r1.match(y)
            if g1 is None or g2 is None:
                return False
            x, y = float(g1.group(2)), float(g2.group(2))
            return myokit.float.close(x, y)

        # Test without arguments
        x = m.format_state_derivatives().splitlines()
        y = [
'membrane.V = -84.5286                   dot = -5.68008003798848027e-02', # noqa
'ina.m      = 0.0017                     dot = -4.94961486033834719e-03', # noqa
'ina.h      = 0.9832                     dot =  9.02025299127830887e-06', # noqa
'ina.j      = 0.995484                   dot = -3.70409866928434243e-04', # noqa
'ica.d      = 3e-06                      dot =  3.68067721821794798e-04', # noqa
'ica.f      = 1.0                        dot = -3.55010150519739432e-07', # noqa
'ik.x       = 0.0057                     dot = -2.04613933160084307e-07', # noqa
'ica.Ca_i   = 0.0002                     dot = -6.99430692442154227e-06'  # noqa
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test with state argument
        state1 = [1, 2, 3, 4, 5, 6, 7, 8]
        state1[2] = 536.46745856785678567845745637
        x = m.format_state_derivatives(state1).splitlines()
        y = [
'membrane.V = 1                          dot =  1.90853168050245158e+07', # noqa
'ina.m      = 2                          dot = -1.56738349674489310e+01', # noqa
'ina.h      =  5.36467458567856738e+02   dot = -3.05729251015767022e+03', # noqa
'ina.j      = 4                          dot = -1.15731427949362953e+00', # noqa
'ica.d      = 5                          dot = -1.85001944916516836e-01', # noqa
'ica.f      = 6                          dot = -2.15435819790876573e-02', # noqa
'ik.x       = 7                          dot = -1.25154369264425316e-02', # noqa
'ica.Ca_i   = 8                          dot = -5.63431267451130036e-01', # noqa                                       ^ ^^    ^ ---------   ^
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

        # Test with invalid state argument
        self.assertRaisesRegex(
            ValueError, r'sequence of \(8\)',
            m.format_state_derivatives, [1, 2, 3])

        # Test with state and precision argument
        # Ignoring some of the middle digits, as they differ on some (but not
        # all!) CI builds.
        out = m.format_state_derivatives(
            state1, precision=myokit.SINGLE_PRECISION).splitlines()
        self.assertEqual(len(out), 8)
        self.assertEqual(out[0][:15], 'membrane.V = 1 ')
        self.assertEqual(out[1][:15], 'ina.m      = 2 ')
        self.assertEqual(out[2][:29], 'ina.h      =  5.364674586e+02')
        self.assertEqual(out[3][:15], 'ina.j      = 4 ')
        self.assertEqual(out[4][:15], 'ica.d      = 5 ')
        self.assertEqual(out[5][:15], 'ica.f      = 6 ')
        self.assertEqual(out[6][:15], 'ik.x       = 7 ')
        self.assertEqual(out[7][:15], 'ica.Ca_i   = 8 ')
        out = [x[x.index('dot') + 6:] for x in out]
        self.assertEqual(out[0][:8], ' 1.90853')
        self.assertEqual(out[1][:8], '-1.56738')
        self.assertEqual(out[2][:8], '-3.05729')
        self.assertEqual(out[3][:8], '-1.15731')
        self.assertEqual(out[4][:8], '-1.85001')
        self.assertEqual(out[5][:8], '-2.15435')
        self.assertEqual(out[6][:8], '-1.25154')
        self.assertEqual(out[7][:8], '-5.63431')
        self.assertEqual(out[0][12:], 'e+07')
        self.assertEqual(out[1][12:], 'e+01')
        self.assertEqual(out[2][12:], 'e+03')
        self.assertEqual(out[3][12:], 'e+00')
        self.assertEqual(out[4][12:], 'e-01')
        self.assertEqual(out[5][12:], 'e-02')
        self.assertEqual(out[6][12:], 'e-02')
        self.assertEqual(out[7][12:], 'e-01')

        # Test with derivs argument
        self.assertEqual(
            m.format_state_derivatives(
                [1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1]),
            'membrane.V = 1                          dot = 8\n'
            'ina.m      = 2                          dot = 7\n'
            'ina.h      = 3                          dot = 6\n'
            'ina.j      = 4                          dot = 5\n'
            'ica.d      = 5                          dot = 4\n'
            'ica.f      = 6                          dot = 3\n'
            'ik.x       = 7                          dot = 2\n'
            'ica.Ca_i   = 8                          dot = 1'
        )

        # Test with invalid derivs argument
        self.assertRaisesRegex(
            ValueError, r'sequence of \(8\)', m.format_state_derivatives,
            [1, 2, 3, 4, 5, 6, 7, 8], [1, 2, 3])

        # Test without expressions in the initial state (these will be
        # evaluated)
        m.get('ina.m').set_initial_value('sqrt(0.25)')
        m.get('ina.h').set_initial_value('ina.gNa / 20')
        x = m.format_state_derivatives().splitlines()
        y = [
'membrane.V = -84.5286                   dot =  2.46843981754470434e+02', # noqa
'ina.m      = 0.5                        dot = -8.68225658924664856e+01', # noqa
'ina.h      = 0.8                        dot =  4.89677127030544446e-02', # noqa
'ina.j      = 0.995484                   dot = -3.70409866928434243e-04', # noqa
'ica.d      = 3e-06                      dot =  3.68067721821794798e-04', # noqa
'ica.f      = 1.0                        dot = -3.55010150519739432e-07', # noqa
'ik.x       = 0.0057                     dot = -2.04613933160084307e-07', # noqa
'ica.Ca_i   = 0.0002                     dot = -6.99430692442154227e-06'  # noqa
        ]
        for a, b in zip(x, y):
            self.assertTrue(almost_equal(a, b))
        self.assertEqual(len(x), len(y))

    def test_get(self):
        # Test Model.get().

        m = myokit.load_model('example')

        # Get by name
        v = m.get('membrane.V')
        self.assertEqual(v.qname(), 'membrane.V')
        self.assertIsInstance(v, myokit.Variable)

        # Get by variable ref (useful for handling unknown input type)
        w = m.get(v)
        self.assertIs(w, v)

        # Get by variable from another model
        m2 = m.clone()
        self.assertRaisesRegex(ValueError, 'different model', m2.get, v)

        # Get nested
        a = m.get('ina.m.alpha')
        self.assertEqual(a.qname(), 'ina.m.alpha')
        self.assertIsInstance(a, myokit.Variable)

        # Get component
        c = m.get('membrane')
        self.assertEqual(c.qname(), 'membrane')
        self.assertIsInstance(c, myokit.Component)

        # Get with filter
        a = m.get('ina.m.alpha', myokit.Variable)
        self.assertEqual(a.qname(), 'ina.m.alpha')
        self.assertIsInstance(a, myokit.Variable)
        self.assertRaises(KeyError, m.get, 'ina.m.alpha', myokit.Component)
        self.assertRaises(KeyError, m.get, 'ina', myokit.Variable)
        m.get('ina', myokit.Component)

        # Get non-existent
        self.assertRaises(KeyError, m.get, 'membrane.bert')
        self.assertRaises(KeyError, m.get, 'bert.bert')

    def test_has_variables(self):
        # Test VarProvider.has_variables (and VarProvider.variables)

        m = myokit.Model()
        z = m.add_component('z')
        self.assertFalse(m.has_variables())

        # Constant
        a = z.add_variable('a')
        a.set_rhs(0)
        self.assertTrue(m.has_variables())
        self.assertTrue(m.has_variables(const=True))
        self.assertFalse(m.has_variables(const=False))

        # State
        self.assertFalse(m.has_variables(state=True))
        b = z.add_variable('b')
        b.set_rhs(1)
        b.promote(0.2)
        self.assertTrue(m.has_variables(state=True))
        self.assertFalse(m.has_variables(const=False, state=False))

        # Inter and deep
        self.assertFalse(m.has_variables(inter=True))
        self.assertFalse(m.has_variables(inter=True, deep=True))
        c = b.add_variable('c')
        c.set_rhs('b * 2')
        b.set_rhs('1 + c')
        self.assertFalse(m.has_variables(inter=True))
        self.assertTrue(m.has_variables(inter=True, deep=True))

        # Bound
        self.assertFalse(m.has_variables(bound=True))
        t = z.add_variable('t')
        t.set_rhs(0)
        t.set_binding('time')
        self.assertTrue(m.has_variables(bound=True))
        self.assertFalse(
            m.has_variables(const=False, state=False, bound=False))

    def test_import_component(self):
        # Test :meth: 'import_component()'.

        # Source model, to import stuff from
        ms = myokit.parse_model('''
            [[model]]
            n.c = 1 + n.p
            n.b = sqrt(0.25)
            p.b = 0.2
            q.e = 0.2
            y.e = 0.2
            x.a = 0.2

            [e]
            t = 0 [s] bind time
                in [s]
            g = 0
                label this_is_g
            h = 4

            # Independent (except for time)
            [p]
            a = 1 [1/s]
                in [1/s]
            # Comments should be stripped out during parsing, so this is OK.
            dot(b) = a * c
                in [m]
                c = b * 2
                    in [m]

            # Requires mapping, has aliases and an unused alias
            [q]
            use p.a, p.b
            use e.h
            dot(e) = d * b
                in [m*A]
            f = 3 * dot(b)
                in [m/s]
            d = 0.2 [A] * a
                in [A/s]

            # Two components to import at the same time
            # (independent of the rest of the model)
            [x]
            use y.e
            dot(a) = c * e
                in [m*A]
            b = 3 * dot(e)
                in [m/s]
            c = 0.2 [A] * d
                in [A/s]
            d = 1 [1/s]
                in [1/s]

            [y]
            use x.d
            dot(e) = d * sub_e
                in [m]
                sub_e = 2 * e
                    in [m]

            # Another group component but this isn't independent
            [z]
            use x.d
            use y.e
            use e.h
            f = d * e * h
                in [m/s]

            # A component with expressions in initial states
            [n]
            a = 1
            dot(b) = 1 [1/s]
            dot(c) = 2 [1/s]
            p = 0.1
        ''')
        ms.validate()
        ms.check_units(myokit.UNIT_STRICT)

        # Make copy of ms, to ensure nothing is altered in original model
        ms_unaltered = ms.clone()

        # Import independent component into empty model
        m1 = myokit.Model()
        m1.import_component(ms['p'])
        self.assertTrue(m1.has_component('p'))
        self.assertFalse(m1['p'] is ms['p'])
        self.assertEqual(m1['p'].code(), ms['p'].code())
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import a second time, without renaming
        m1.import_component, ms['p']  # Check errors happen before changes
        m1_unaltered = m1.clone()
        self.assertRaises(myokit.DuplicateName, m1.import_component, ms['p'])
        self.assertTrue(m1.is_similar(m1_unaltered, True))

        # Import a second time, and rename
        m1.import_component(ms['p'], new_name='p2')
        self.assertTrue(m1.has_component('p2'))
        self.assertFalse(m1['p2'] is ms['p'])
        self.assertFalse(m1['p2'] is m1['p'])
        cs = '\n'.join((ms['p'].code().splitlines())[1:])
        c1 = '\n'.join((m1['p2'].code().splitlines())[1:])
        self.assertEqual(cs, c1)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import independent component with labels and bindings
        m1.import_component(ms['e'])
        self.assertTrue(m1.has_component('e'))
        self.assertFalse(m1['e'] is ms['e'])
        self.assertEqual(m1['e'].code(), ms['e'].code())
        self.assertEqual(m1.label('this_is_g'), m1.get('e.g'))
        self.assertEqual(m1.binding('time'), m1.get('e.t'))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Now that it has a time variable, m1 should be valid
        m1.validate()
        self.assertTrue(m1.is_valid())

        # Re-importing e is not allowed, as it leads to double bindings and/or
        # labels
        m1_unaltered = m1.clone()
        ms.time().set_binding(None)
        self.assertRaisesRegex(
            myokit.InvalidLabelError, 'label "this_is_g"',
            m1.import_component, ms['e'], new_name='dinosaur')
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        ms.get('e.t').set_binding('time')
        ms.label('this_is_g').set_label(None)
        self.assertRaisesRegex(
            myokit.InvalidBindingError, 'binding "time"',
            m1.import_component, ms['e'], new_name='hello')
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        ms = ms_unaltered.clone()

        # Import r, using a custom variable mapping
        var_map = {'p.a': 'p.a', 'p.b': 'p2.b', 'e.h': 'e.h'}
        m1.import_component(ms['q'], var_map=var_map)
        self.assertTrue(m1.has_component('q'))
        self.assertFalse(m1['q'] is ms['q'])
        self.assertEqual(m1['q'].alias('a'), m1.get('p.a'))
        self.assertEqual(m1['q'].alias('b'), m1.get('p2.b'))
        self.assertEqual(m1['q'].alias('h'), m1.get('e.h'))
        cs = '\n'.join((ms['p'].code().splitlines())[4:])
        c1 = '\n'.join((m1['p2'].code().splitlines())[4:])
        self.assertEqual(cs, c1)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import r, using a label for p.b
        ms.get('p.b').set_label('this_is_b')
        m1.get('p2.b').set_label('this_is_b')
        var_map = {'p.a': 'p.a', 'e.h': 'e.h'}
        m1.import_component(ms['q'], var_map=var_map, new_name='q2')
        self.assertTrue(m1.has_component('q2'))
        self.assertFalse(m1['q2'] is ms['q'])
        self.assertFalse(m1['q2'] is m1['q'])
        self.assertEqual(m1['q2'].alias('a'), m1.get('p.a'))
        self.assertEqual(m1['q2'].alias('b'), m1.get('p2.b'))
        cs = '\n'.join((ms['p'].code().splitlines())[3:])
        c1 = '\n'.join((m1['p2'].code().splitlines())[3:])
        self.assertEqual(cs, c1)
        ms.get('p.b').set_label(None)
        m1.get('p2.b').set_label(None)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import r, using a binding for e.h
        ms.get('e.h').set_binding('this_is_h')
        m1.get('e.h').set_binding('this_is_h')
        var_map = {'p.a': 'p.a', 'p.b': 'p2.b'}
        m1.import_component(ms['q'], var_map=var_map, new_name='q3')
        self.assertTrue(m1.has_component('q3'))
        self.assertFalse(m1['q3'] is ms['q'])
        self.assertFalse(m1['q3'] is m1['q'])
        self.assertEqual(m1['q3'].alias('a'), m1.get('p.a'))
        self.assertEqual(m1['q3'].alias('b'), m1.get('p2.b'))
        cs = '\n'.join((ms['p'].code().splitlines())[3:])
        c1 = '\n'.join((m1['p2'].code().splitlines())[3:])
        self.assertEqual(cs, c1)
        ms.get('e.h').set_binding(None)
        m1.get('e.h').set_binding(None)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import r, using (partial) name mapping
        var_map = {'p.a': 'p2.a'}
        m1.import_component(
            ms['q'], var_map=var_map, new_name='q4', allow_name_mapping=True)
        self.assertTrue(m1.has_component('q3'))
        self.assertFalse(m1['q3'] is ms['q'])
        self.assertFalse(m1['q4'] is m1['q'])
        self.assertEqual(m1['q4'].alias('a'), m1.get('p2.a'))
        cs = '\n'.join((ms['p'].code().splitlines())[2:])
        c1 = '\n'.join((m1['p2'].code().splitlines())[2:])
        self.assertEqual(cs, c1)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import multiple components
        component_list = [ms['x'], ms['y']]
        m1.import_component(component_list)
        self.assertTrue(m1.has_component('x'))
        self.assertTrue(m1.has_component('y'))
        self.assertFalse(m1['x'] is ms['x'])
        self.assertFalse(m1['y'] is ms['y'])
        self.assertEqual(m1['x'].code(), ms['x'].code())
        self.assertEqual(m1['y'].code(), ms['y'].code())
        self.assertTrue(ms.is_similar(ms_unaltered, True))
        # Check that state order is preserved
        self.assertLess(m1.get('y.e').index(), m1.get('x.a').index())

        # Import multiple components and rename
        component_list = [ms['x'], ms['y']]
        m1.import_component(component_list, new_name=['xx', 'yy'])
        self.assertTrue(m1.has_component('xx'))
        self.assertTrue(m1.has_component('yy'))
        self.assertFalse(m1['xx'] is ms['x'])
        self.assertFalse(m1['yy'] is ms['y'])
        self.assertTrue(ms.is_similar(ms_unaltered, True))
        # Check that state order is preserved
        self.assertLess(m1.get('yy.e').index(), m1.get('xx.a').index())

        # Import 1 component in list
        m1.import_component([ms['p']], new_name='p3')
        self.assertTrue(m1.has_component('p3'))
        self.assertFalse(m1['p3'] is ms['p'])
        self.assertFalse(m1['p3'] is m1['p'])
        cs = '\n'.join((ms['p'].code().splitlines())[1:])
        c1 = '\n'.join((m1['p3'].code().splitlines())[1:])
        self.assertEqual(cs, c1)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import state variables with expressions in their initial values
        m1_unaltered = m1.clone()
        m1.import_component(ms['n'])
        self.assertEqual(m1.get('n.b').initial_value().code(), 'sqrt(0.25)')
        self.assertEqual(
            m1.get('n.c').initial_value(),
            myokit.Plus(myokit.Number(1), myokit.Name(m1.get('n.p'))))
        # Check that state order is preserved
        self.assertLess(m1.get('n.c').index(), m1.get('n.b').index())
        m1 = m1_unaltered

        # ...including references to parameters in other components
        ms2 = ms.clone()
        ms2.get('n.c').set_initial_value('1 + e.h')
        ms2_unaltered = ms2.clone()
        self.assertRaises(
            myokit.VariableMappingError, m1.import_component, ms2['n'])
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms2.is_similar(ms2_unaltered, True))

        # Try and fail to import r without a mapping
        m1_unaltered = m1.clone()
        self.assertRaises(
            myokit.VariableMappingError,
            m1.import_component, ms['q'], new_name='q9')
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # With an invalid mapping
        self.assertRaisesRegex(
            TypeError, 'dict or None',
            m1.import_component, ms['q'], new_name='q9', var_map=[3])
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertRaisesRegex(
            TypeError, 'objects or fully qualified',
            m1.import_component, ms['q'], new_name='q9',
            var_map={'p.a': 123, 'p.b': 'p.b', 'e.h': 'e.h'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertRaisesRegex(
            TypeError, 'objects or fully qualified',
            m1.import_component, ms['q'], new_name='q9',
            var_map={345: 'p.a', 'p.b': 'p.b', 'e.h': 'e.h'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'Multiple variables map',
            m1.import_component, ms['q'], new_name='q9',
            var_map={'p.a': 'p.a', 'p.b': 'p.a', 'e.h': 'e.h'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # With variables from another model or that don't exist
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'was not found in this model',
            m1.import_component, ms['q'], new_name='q9',
            var_map={'p.a': 'one.two', 'p.b': 'p.b', 'e.h': 'e.h'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        self.assertRaisesRegex(
            myokit.VariableMappingError, 'was not found in the source model',
            m1.import_component, ms['q'], new_name='q9',
            var_map={'ppp.aaa': 'p.a', 'p.b': 'p.b', 'e.h': 'e.h'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        v = myokit.Model('a').add_component('p').add_variable('b')
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'is not part of this model',
            m1.import_component, ms['q'], new_name='q9',
            var_map={'p.a': v, 'p.b': 'p.b'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        self.assertRaisesRegex(
            myokit.VariableMappingError, 'is not part of the source model',
            m1.import_component, ms['q'], new_name='q9',
            var_map={v: 'p.a', 'p.b': 'p.b', 'e.h': 'e.h'})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Incomplete mapping
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'cannot be mapped',
            m1.import_component, ms['q'], new_name='q9', var_map={})
        self.assertTrue(m1.is_similar(m1_unaltered, True))
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Import something that's not a component
        self.assertRaisesRegex(
            TypeError, 'myokit.Component',
            m1.import_component, 'q')

        self.assertRaisesRegex(
            TypeError, 'myokit.Component',
            m1.import_component, [ms['q'], 'q'])

        # Import your own components
        self.assertRaisesRegex(
            ValueError, 'part of this model',
            m1.import_component, m1['p'], new_name='abc')

        # new_name is not string or list of correct length
        self.assertRaisesRegex(
            TypeError, 'new_name must be',
            m1.import_component, m1['p'], new_name=1)

        self.assertRaisesRegex(
            TypeError, 'new_name must be',
            m1.import_component, [m1['p'], m1['q']], new_name='abs')

        self.assertRaisesRegex(
            TypeError, 'new_name must be',
            m1.import_component, [m1['p'], m1['q']], new_name=['abs', 1])

        self.assertRaisesRegex(
            TypeError, 'new_name must be',
            m1.import_component, [m1['p'], m1['q']], new_name=['abs'])

        # Multiple imported components must be all from the same model
        m2 = myokit.parse_model('''
            [[model]]
            p.b = 0.2

            # Independent (except for time)
            [p]
            t = 0 [s] bind time
                in [s]
            a = 1 [m]
                in [m]
            # Comments should be stripped out during parsing, so this is OK.
            dot(b) = 2 * a
                in [m*s]

        ''')
        m2.validate()
        m2.check_units(myokit.UNIT_STRICT)

        self.assertRaisesRegex(
            ValueError, 'must be from the same model',
            m1.import_component, [m1['p'], m2['p']])

    def test_import_component_units(self):
        # Test :meth: 'import_component()' with unit conversion.

        # Source model, to import stuff from
        ms = myokit.parse_model('''
            [[model]]
            p.c = 0.1
            r.f = 0.2
            x.a = 0.2
            y.e = 0.2

            [e]
            t = 0 [s] bind time
                in [s]

            # Independent (except for time)
            [p]
            a = 1 [m] in [m]
            b = 10 [s] in [s]
            dot(c) = 0.2 [m/s]
                in [m]

            # Requires mapping, doesn't have states but has dot() reference
            [q]
            d = 3.2 [m/s] + dot(p.c)
                in [m/s]
            e = p.a + p.c + 0.1 [m]
                in [m]

            # Requires mapping, has a new state
            [r]
            use p.a, p.b
            dot(f) = a / b
                in [m]

            # Two components to import at the same time
            # (independent of the rest of the model)
            [x]
            use y.e
            dot(a) = c * e
                in [m*A]
            b = 3 * dot(e)
                in [m/s]
            c = 0.2 [A] * d
                in [A/s]
            d = 1 [1/s]
                in [1/s]

            [y]
            use x.d
            dot(e) = d * sub_e
                in [m]
                sub_e = 2 * e
                    in [m]

            # No time units used
            [s]
            x = 3 [kg] in [kg]
        ''')
        ms.validate()
        ms.check_units(myokit.UNIT_STRICT)

        # Make copy of ms, to ensure nothing is altered in original model
        ms_unaltered = ms.clone()

        # Conversion helpers
        vm = {'p.a': 'p.a', 'p.b': 'p.b', 'p.c': 'p.c'}
        s2ms = myokit.Number(1000, myokit.units.ms / myokit.units.s)
        mm2m = myokit.Number(0.001, myokit.units.m / myokit.units.mm)

        # Target model to import stuff into, with different time units
        m1 = myokit.parse_model('''
            [[model]]

            [e]
            t = 0 [ms] bind time
                in [ms]
        ''')
        m1.validate()
        m1.check_units(myokit.UNIT_STRICT)

        # Import p, should be same
        m1.import_component(ms['p'], convert_units=True)

        # Import q, should be same except for dot() expression
        m1.import_component(ms['q'], var_map=vm, convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))
        self.assertIn('q', m1)
        self.assertEqual(len(m1['q']), 2)
        self.assertEqual(m1.get('q.e').code(), ms.get('q.e').code())
        self.assertEqual(m1.get('q.d').unit(), ms.get('q.d').unit())
        self.assertEqual(
            m1.get('q.d').rhs().code(),
            myokit.Plus(myokit.Number(3.2, 'm/s'), myokit.Multiply(
                myokit.Derivative(myokit.Name(m1.get('p.c'))), s2ms)
            ).code())

        # Import r, should be same except for the state's RHS
        m1.import_component(ms['r'], var_map=vm, convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))
        self.assertIn('r', m1)
        self.assertEqual(len(m1['r']), 1)
        self.assertIn('f', m1['r'])
        self.assertTrue(m1.get('r.f').is_state())
        self.assertEqual(m1.get('r.f').unit(), ms.get('r.f').unit())
        self.assertEqual(
            m1.get('r.f').initial_value(), ms.get('r.f').initial_value())
        self.assertEqual(
            m1.get('r.f').rhs().code(),
            myokit.Multiply(ms.get('r.f').rhs(), s2ms).code())

        # Import multiple components
        m1.import_component([ms['x'], ms['y']], convert_units=True)
        self.assertTrue(m1.has_component('x'))
        self.assertTrue(m1.has_component('y'))
        self.assertFalse(m1['x'] is ms['x'])
        self.assertFalse(m1['y'] is ms['y'])
        self.assertEqual(m1.get('x.a').unit(), ms.get('x.a').unit())
        self.assertEqual(
            m1.get('x.a').initial_value(), ms.get('x.a').initial_value())
        self.assertEqual(
            m1.get('x.a').rhs().code(),
            myokit.Multiply(ms.get('x.a').rhs(), s2ms).code())
        self.assertEqual(
            m1.get('y.e').initial_value(), ms.get('y.e').initial_value())
        self.assertEqual(
            m1.get('y.e').rhs().code(),
            myokit.Multiply(ms.get('y.e').rhs(), s2ms).code())

        # Target model with different space units
        m1 = myokit.parse_model('''
            [[model]]
            p.c = 0.1

            [e]
            time = 0 [s] bind time
                in [s]

            [p]
            a = 1 [mm] in [mm]
            b = 10 [s] in [s]
            dot(c) = 0.2 [mm/s]
                in [mm]
        ''')
        m1.validate()
        m1.check_units(myokit.UNIT_STRICT)

        # Import q, converting p.a and dot(p.c)
        m1.import_component(ms['q'], var_map=vm, convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))
        self.assertIn('q', m1)
        self.assertEqual(len(m1['q']), 2)
        self.assertEqual(m1.get('q.d').unit(), ms.get('q.d').unit())
        self.assertEqual(
            m1.get('q.d').rhs().code(),
            myokit.Plus(
                myokit.Number(3.2, 'm/s'),
                myokit.Multiply(
                    myokit.Derivative(myokit.Name(m1.get('p.c'))), mm2m)
            ).code())
        self.assertEqual(m1.get('q.e').unit(), ms.get('q.e').unit())
        self.assertEqual(
            m1.get('q.e').rhs().code(),
            myokit.Plus(myokit.Plus(
                myokit.Multiply(myokit.Name(m1.get('p.a')), mm2m),
                myokit.Multiply(myokit.Name(m1.get('p.c')), mm2m)),
                myokit.Number(0.1, 'm')).code())

        # Target model with different time and space units
        m1 = myokit.parse_model('''
            [[model]]
            p.c = 0.1

            [e]
            time = 0 [ms] bind time
                in [ms]

            [p]
            a = 1 [mm] in [mm]
            b = 10 [ms] in [ms]
            dot(c) = 0.2 [mm/ms]
                in [mm]
        ''')
        m1.validate()
        m1.check_units(myokit.UNIT_STRICT)

        # Import q and p, converting p.a and dot(p.c)
        m1.import_component(ms['q'], var_map=vm, convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))
        self.assertIn('q', m1)
        self.assertEqual(len(m1['q']), 2)
        self.assertEqual(m1.get('q.d').unit(), ms.get('q.d').unit())
        self.assertEqual(
            m1.get('q.d').rhs().code(),
            myokit.Plus(
                myokit.Number(3.2, 'm/s'),
                myokit.Derivative(myokit.Name(m1.get('p.c')))
            ).code())
        self.assertEqual(m1.get('q.e').unit(), ms.get('q.e').unit())
        self.assertEqual(
            m1.get('q.e').rhs().code(),
            myokit.Plus(myokit.Plus(
                myokit.Multiply(myokit.Name(m1.get('p.a')), mm2m),
                myokit.Multiply(myokit.Name(m1.get('p.c')), mm2m)),
                myokit.Number(0.1, 'm')).code())

        # Attempt import with incompatible time units 1:
        # Component that defines a derivative in m/s
        m1 = myokit.parse_model('''
            [[model]]

            [e]
            time = 0 [A] bind time
                in [A]
        ''')
        m1.validate()
        m1.check_units(myokit.UNIT_STRICT)
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'time variables',
            m1.import_component, ms['p'], convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Attempt import with incompatible time units 1:
        # Component that uses a dot expression
        m1 = myokit.parse_model('''
            [[model]]
            p.c = 0

            [e]
            time = 0 [A] bind time
                in [A]

            [p]
            a = 1 [m] in [m]
            dot(c) = 0.2 [m/A]
                in [m]
        ''')
        m1.validate()
        m1.check_units(myokit.UNIT_STRICT)
        vm = {'p.a': 'p.a', 'p.c': 'p.c'}
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'time variables',
            m1.import_component, ms['q'], var_map=vm, convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

        # Attempt import with incompatible mapped (space) units
        m1 = myokit.parse_model('''
            [[model]]
            p.c = 0

            [e]
            time = 0 [s] bind time
                in [s]

            [p]
            a = 1 [V] in [V]
            b = 10 [s] in [s]
            dot(c) = 0.2 [V/s]
                in [V]
        ''')
        m1.validate()
        m1.check_units(myokit.UNIT_STRICT)
        vm = {'p.a': 'p.a', 'p.b': 'p.b', 'p.c': 'p.c'}
        self.assertRaisesRegex(
            myokit.VariableMappingError, 'Unable to convert',
            m1.import_component, ms['q'], var_map=vm, convert_units=True)
        self.assertTrue(ms.is_similar(ms_unaltered, True))

    def test_initial_values(self):
        # Tests :meth:`Model.initial_values`.

        m = myokit.parse_model('''
            [[model]]
            c.x = 2
            c.y = 1 + 2
            c.z = 2 * c.p

            [c]
            t = 0 bind time
            dot(x) = 0
            dot(y) = 0
            dot(z) = 0
            p = 3
            q = 4
        ''')

        # Test expression version
        p = m.get('c.p')
        x0 = m.initial_values()
        self.assertEqual(x0[0], myokit.Number(2))
        self.assertEqual(
            x0[1], myokit.Plus(myokit.Number(1), myokit.Number(2)))
        self.assertEqual(x0[2], myokit.Multiply(myokit.Number(2), p.lhs()))

        # Test float version
        self.assertEqual(m.initial_values(True), [2, 3, 6])

        # Test deprecated alias
        with WarningCollector() as w:
            self.assertEqual(m.state(), m.initial_values(True))
        self.assertIn('deprecated', w.text())

        # Test cycles are detected before evaluation
        p.set_rhs('1 / q')
        m.get('c.q').set_rhs('1 + p')
        m.initial_values()  # No evaluation, so no error
        self.assertRaises(
            myokit.CyclicalDependencyError, m.initial_values, as_floats=True)

    def test_inits(self):
        # Tests :meth:`Model.inits`

        m = myokit.parse_model('''
            [[model]]
            c.x = 2
            c.y = 1 + 2
            c.z = 2 * c.p

            [c]
            t = 0 bind time
            dot(x) = 0
            dot(y) = 0
            dot(z) = 0
            p = 3
        ''')
        with WarningCollector() as w:
            eqs = m.inits()
        self.assertIn('deprecated', w.text())

        E, N, M = myokit.Equation, myokit.Number, myokit.Multiply
        self.assertEqual(list(eqs), [
            E(myokit.Name(m.get('c.x')), N(2)),
            E(myokit.Name(m.get('c.y')), myokit.Plus(N(1), N(2))),
            E(myokit.Name(m.get('c.z')), M(N(2), m.get('c.p').lhs()))
        ])

    def test_is_similar(self):
        # Check that equality takes both code() and unames into account

        # Test without custom reserved names
        m1 = myokit.load_model('example')
        m2 = m1.clone()
        self.assertIsInstance(m2, myokit.Model)
        self.assertFalse(m1 is m2)
        self.assertNotEqual(m1, m2)
        self.assertNotEqual(m2, m1)
        self.assertTrue(m1.is_similar(m2, False))
        self.assertTrue(m1.is_similar(m2, True))
        self.assertTrue(m2.is_similar(m1, False))
        self.assertTrue(m2.is_similar(m1, True))
        self.assertTrue(m1.is_similar(m1, True))
        self.assertTrue(m2.is_similar(m2, False))

        # Test with none-model
        self.assertFalse(m1.is_similar(None))
        self.assertFalse(m1.is_similar(m1.code()))

        # Add reserved names
        m1.reserve_unique_names('bertie')
        self.assertFalse(m1.is_similar(m2))
        m1.reserve_unique_names('clair')
        self.assertFalse(m1.is_similar(m2))
        m2.reserve_unique_names('clair', 'bertie')
        self.assertTrue(m1.is_similar(m2))

        # Add reserved name prefixes
        m1.reserve_unique_name_prefix('aa', 'bb')
        m1.reserve_unique_name_prefix('cc', 'dd')
        self.assertFalse(m1.is_similar(m2))
        m2.reserve_unique_name_prefix('aa', 'bb')
        m2.reserve_unique_name_prefix('cc', 'ee')
        self.assertFalse(m1.is_similar(m2))
        m2.reserve_unique_name_prefix('cc', 'dd')
        self.assertTrue(m1.is_similar(m2))

    def test_item_at_text_position(self):
        # Test :meth:`Model.item_at_text_position()`.

        text = [
            '[[model]]',        # 1
            'c.x = 0',          # 2
            '',                 # 3
            '[e]',              # 4
            't = 0 bind time',
            '',
            '[c]',
            'desc: This is a test component',
            'dot(x) = (10 - x) / y',
            'y = 5 + y1',
            '    y1 = 3',
            ''
        ]
        model = myokit.parse_model(text)
        e = model.get('e')
        t = model.get('e.t')
        c = model.get('c')
        x = model.get('c.x')
        y = model.get('c.y')
        y1 = model.get('c.y.y1')

        def check(line, char, var):
            tv = model.item_at_text_position(line, char)
            if var is None:
                self.assertIsNone(tv)
            else:
                token, var2 = tv
                self.assertIsNotNone(var2)
                self.assertEqual(var.qname(), var2.qname())

        # It doesn't work on initial conditions
        check(1, 0, None)
        check(1, 1, None)
        check(1, 2, None)

        # Find the component e and its variable
        check(4, 0, None)
        check(4, 1, e)
        check(4, 2, None)
        check(5, 0, t)
        check(5, 1, None)

        # Find the component c and its variables
        check(7, 1, c)
        check(9, 4, x)
        check(10, 0, y)
        check(11, 4, y1)

    def test_label(self):
        # Tests setting and getting of labels

        # Note that set_label() is part of Variable, so not tested here
        m = myokit.Model()
        c = m.add_component('c')
        x = c.add_variable('x')
        x.set_label('hello')
        y = c.add_variable('y')
        y.set_label('goodbye')
        z = c.add_variable('z')
        z.set_label('x')

        # Test label()
        self.assertEqual(m.label('goodbye'), y)
        self.assertEqual(m.label('hello'), x)
        self.assertEqual(m.label('x'), z)
        self.assertEqual(m.label('y'), None)

        # Test labels()
        labels = dict(m.labels())
        self.assertEqual(labels['goodbye'], y)
        self.assertEqual(labels['hello'], x)
        self.assertEqual(labels['x'], z)
        self.assertFalse('y' in labels)

        # Test labelx()
        self.assertEqual(m.labelx('goodbye'), y)
        self.assertEqual(m.labelx('hello'), x)
        self.assertEqual(m.labelx('x'), z)
        self.assertRaisesRegex(
            myokit.IncompatibleModelError,
            'No variable found with label "y"',
            m.labelx, 'y')

        # Test incompatible model error includes name if possible
        self.assertRaisesRegex(
            myokit.IncompatibleModelError,
            'Incompatible model: No variable found',
            m.labelx, 'y')
        m.set_name('Bert')
        self.assertRaisesRegex(
            myokit.IncompatibleModelError,
            'Incompatible model <Bert>: No variable found',
            m.labelx, 'y')

    def test_labels(self):
        # Test setting labels and :meth:`Model.labels()`.

        # Test set_label() and labels()
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        v = c.add_variable('v')
        v.set_rhs('3 - v')
        v.set_label('membrane_potential')
        w = c.add_variable('w')
        w.set_rhs(1)
        x = c.add_variable('x')
        x.set_rhs(1)
        labels = list(m.labels())
        self.assertEqual(len(labels), 1)
        self.assertEqual(labels[0][0], 'membrane_potential')
        self.assertEqual(labels[0][1], v)

        # Can't have two labels
        self.assertRaisesRegex(
            myokit.InvalidLabelError, 'already has a label', v.set_label,
            'bert')

        # No two variables can have the same label
        self.assertRaisesRegex(
            myokit.InvalidLabelError, 'already in use', w.set_label,
            'membrane_potential')

        # Labels can't overlap with bindings
        self.assertRaisesRegex(
            myokit.InvalidLabelError, 'in use as a binding', w.set_label,
            'time')

    def test_load_save_state(self):
        # Test :meth:`Model.save_state()` and :meth:`Model.load_state()`.

        m = myokit.load_model('example')
        s1 = m.initial_values(True)
        with TemporaryDirectory() as d:
            path = d.path('state.csv')
            with WarningCollector() as w:
                m.save_state(path)
            self.assertIn('deprecated', w.text())
            self.assertEqual(m.initial_values(True), s1)
            sx = list(s1)
            sx[0] = 10
            m.set_initial_values(sx)
            self.assertNotEqual(m.initial_values(True), s1)
            with WarningCollector() as w:
                m.load_state(path)
            self.assertIn('deprecated', w.text())
            self.assertEqual(m.initial_values(True), s1)

    def test_map_to_state(self):
        # Test :meth:`Model.map_to_state()`.

        # Create test model
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        v = c.add_variable('v')
        v.set_rhs('3 - v')
        v.promote(0)
        w = c.add_variable('w')
        w.set_rhs('1 - w')
        w.promote(0)

        # List of numbers
        x = m.map_to_state([1, 2])
        self.assertEqual(x, [1.0, 2.0])

        # Wrong size list
        self.assertRaisesRegex(
            ValueError, 'Wrong number', m.map_to_state, [1, 2, 3])

        # String not tested, handled by parse_state.

        # Dict of names
        x = m.map_to_state({'c.v': 2, 'c.w': 3})
        self.assertEqual(x, [2.0, 3.0])

        # Dict of Variables
        x = m.map_to_state({v: 2, 'c.w': 3})
        self.assertEqual(x, [2.0, 3.0])

        # Missing state
        self.assertRaisesRegex(
            ValueError, 'Missing state', m.map_to_state, {v: 2})

    def test_no_rhs_error(self):
        # Test an exception is raised when a variable is missing an rhs.

        m = myokit.Model('LotkaVolterra')
        c0 = m.add_component('c0')
        t = c0.add_variable('time')
        t.set_binding('time')
        self.assertRaises(myokit.MissingRhsError, m.validate)
        t.set_rhs(myokit.Number(0))
        m.validate()
        a = c0.add_variable('test')
        self.assertRaises(myokit.MissingRhsError, m.validate)
        a.set_rhs(myokit.Number(1))
        m.validate()
        b = c0.add_variable('derv')
        b.promote(10)
        self.assertRaises(myokit.MissingRhsError, m.validate)
        b.set_rhs(2)
        m.validate()

    def test_no_time_variable(self):
        # Test an exception is raised if nothing is bound to time.

        m = myokit.Model('LotkaVolterra')
        c0 = m.add_component('c0')
        t = c0.add_variable('time')
        t.set_rhs(myokit.Number(0))
        self.assertRaises(myokit.MissingTimeVariableError, m.validate)

    def test_name(self):
        # Test :meth:`Model.set_name(name)`.

        m = myokit.Model()
        self.assertIsNone(m.name())
        m.set_name('ernie')
        self.assertEqual(m.name(), 'ernie')
        m.set_name(None)
        self.assertIsNone(m.name())
        m.set_name(None)
        self.assertIsNone(m.name())

        m = myokit.Model(name='ernie')
        self.assertEqual(m.name(), 'ernie')
        m.set_name(None)
        self.assertIsNone(m.name())
        m.set_name('bert')
        self.assertEqual(m.name(), 'bert')

    def test_pickling(self):
        # Test pickling and unpickling a model

        # Test model structure
        m1 = myokit.load_model('example')
        m_bytes = pickle.dumps(m1)
        m2 = pickle.loads(m_bytes)
        self.assertFalse(m1 is m2)
        self.assertFalse(m1 == m2)
        self.assertIsInstance(m2, myokit.Model)
        self.assertTrue(m1.is_similar(m2))

        # Test unique names and prefixes (see also test_clone)
        m1.reserve_unique_names('barnard', 'lincoln', 'glasgow')
        m1.reserve_unique_name_prefix('monkey', 'giraffe')
        m1.reserve_unique_name_prefix('ostrich', 'turkey')
        m_bytes = pickle.dumps(m1)
        m2 = pickle.loads(m_bytes)
        self.assertTrue(m1.is_similar(m2, True))

    def test_remove_component(self):
        # Test the removal of a component.

        # Create model
        m = myokit.Model('LotkaVolterra')
        # Simplest case
        X = m.add_component('X')
        self.assertEqual(m.count_components(), 1)
        m.remove_component(X)
        self.assertEqual(m.count_components(), 0)
        self.assertRaises(KeyError, m.remove_component, X)

        # Test if orphaned
        self.assertIsNone(X.parent())

        # Re-adding
        self.assertEqual(m.count_components(), 0)
        X = m.add_component('X')
        self.assertEqual(m.count_components(), 1)

        # With internal variables and string name
        a = X.add_variable('a')
        a.set_rhs(myokit.Number(4))
        b = X.add_variable('b')
        b.set_rhs(myokit.Name(a))
        m.remove_component('X')
        self.assertEqual(m.count_components(), 0)

        # With dependencies from another component
        X = m.add_component('X')
        a = X.add_variable('a')
        a.set_rhs(myokit.Number(45))
        b = X.add_variable('b')
        b.set_rhs(myokit.Name(b))
        Y = m.add_component('Y')
        c = Y.add_variable('c')
        c.set_rhs(myokit.Name(a))
        d = Y.add_variable('d')
        d.set_rhs(myokit.Name(c))
        self.assertEqual(m.count_components(), 2)
        self.assertRaises(myokit.IntegrityError, m.remove_component, X)
        self.assertEqual(m.count_components(), 2)

        # In the right order...
        m.remove_component(Y)
        self.assertEqual(m.count_components(), 1)
        m.remove_component(X)
        self.assertEqual(m.count_components(), 0)

    def test_remove_derivative_references(self):
        # Test the remove_derivative_references() method.

        m0 = myokit.parse_model('''
            [[model]]
            c.x = 0
            c.y = 1

            [e]
            t = 0 bind time

            [c]
            dot(x) = (1 - x) * alpha
                alpha = 3 * beta + dot_x
                beta = exp(-y) + cc
                    cc = 1.23
                dot_x = 3
            dot(y) = (12 - y) / 7
            z = 3 * dot(y)

            [d]
            z = 3 * dot(c.x) / dot(c.y)
            ''')
        m2 = myokit.parse_model('''
            [[model]]
            c.x = 0
            c.y = 1

            [e]
            t = 0 bind time

            [c]
            dot(x) = dot_x_1
            dot_x_1 = (1 - x) * alpha
                alpha = 3 * beta + dot_x
                beta = exp(-y) + cc
                    cc = 1.23
                dot_x = 3
            dot(y) = dot_y
            dot_y = (12 - y) / 7
            z = 3 * dot_y

            [d]
            z = 3 * c.dot_x_1 / c.dot_y
            ''')

        # Remove derivatives from m1
        m1 = m0.clone()
        m1.remove_derivative_references()

        # Assert model matches expected code
        self.assertEqual(m1.code(), m2.code())

        # Assert models both produce the same derivatives
        dy1 = m1.evaluate_derivatives()
        dy2 = m2.evaluate_derivatives()
        self.assertEqual(dy1, dy2)

        # Test time unit is None
        self.assertIsNone(m1.get('c.dot_y').unit())

        # Only one unit set? Then unit is still None
        m1 = m0.clone()
        m1.get('c.y').set_unit('mV')
        m1.remove_derivative_references()
        self.assertIsNone(m1.get('c.dot_y').unit())

        m1 = m0.clone()
        m1.get('e.t').set_unit('ms')
        m1.remove_derivative_references()
        self.assertIsNone(m1.get('c.dot_y').unit())

        # Both units set? Then unit is division of two
        m1 = m0.clone()
        m1.get('e.t').set_unit('ms')
        m1.get('c.y').set_unit('mV')
        m1.remove_derivative_references()
        self.assertEqual(m1.get('c.dot_y').unit(), myokit.parse_unit('mV/ms'))

    def test_reorder_state(self):
        # Test :meth:`Model.reorder_state()`.

        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        v = c.add_variable('v')
        v.set_rhs('3 - v')
        v.promote(0)
        w = c.add_variable('w')
        w.set_rhs(1)
        w.promote(0)
        self.assertEqual(list(m.states()), [v, w])
        m.reorder_state([w, v])
        self.assertEqual(list(m.states()), [w, v])
        m.reorder_state([w, v])
        self.assertEqual(list(m.states()), [w, v])
        m.reorder_state([v, w])
        self.assertEqual(list(m.states()), [v, w])

        # Wrong number of states
        self.assertRaisesRegex(
            ValueError, 'number of entries', m.reorder_state, [v])
        self.assertRaisesRegex(
            ValueError, 'number of entries', m.reorder_state, [v, w, v])

        # Duplicate entries
        self.assertRaisesRegex(
            ValueError, 'Duplicate', m.reorder_state, [v, v])

        # Not a state
        self.assertRaisesRegex(
            ValueError, 'must all be', m.reorder_state, [v, t])

    def test_resolve_interdependent_components(self):
        # Test :meth:`Model.resolve_interdependent_components()`.

        # Create test model
        m = myokit.Model()
        c1 = m.add_component('c1')
        t = c1.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        c2 = m.add_component('c2')
        v = c2.add_variable('v')
        c3 = m.add_component('c3')
        w = c3.add_variable('w')
        x = c3.add_variable('x')
        v.set_rhs(3)
        w.set_rhs(2)
        x.set_rhs(1)

        # Test merge not required
        self.assertEqual(m.has_interdependent_components(), False)

        # Test merge doesn't change model
        self.assertEqual(m.count_components(), 3)
        m.resolve_interdependent_components()
        self.assertEqual(m.count_components(), 3)

        # Create interdependent components
        v.set_rhs('3 - c3.x')
        w.set_rhs('1 - c2.v')

        # Test merge is required
        self.assertTrue(m.has_interdependent_components())

        # Merge
        self.assertEqual(m.count_components(), 3)
        m.resolve_interdependent_components()
        self.assertEqual(m.count_components(), 4)
        m.get('remaining')

        # Test name clash detection
        m = myokit.Model()
        c1 = m.add_component('remaining')
        t = c1.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        c2 = m.add_component('remaining_1')
        v = c2.add_variable('v')
        c3 = m.add_component('c3')
        w = c3.add_variable('w')
        x = c3.add_variable('x')
        v.set_rhs('3 - c3.x')
        w.set_rhs('1 - remaining_1.v')
        x.set_rhs('12')

        self.assertEqual(m.count_components(), 3)
        m.resolve_interdependent_components()
        self.assertEqual(m.count_components(), 4)
        m.get('remaining_2')

    def test_sequence_interface(self):
        # Test the sequence interface implementation
        model = myokit.load_model('example')

        cs = [c for c in model]
        self.assertEqual(cs, list(model.components()))
        self.assertEqual(len(cs), len(model))
        c = model['membrane']
        self.assertEqual(c.name(), 'membrane')

    def test_set_initial_values(self):
        # Tests :meth:`Model.set_initial_values()`.
        m = myokit.load_model('example')
        states = [v.qname() for v in m.states()]
        values = m.initial_values(as_floats=True)

        # Set with floats
        values[0] += 2
        m.set_initial_values(values)
        for val, ref in zip(m.initial_values(True), values):
            self.assertEqual(val, ref)

        # Set with expressions
        values[1] -= 3
        m.set_initial_values([myokit.Number(v) for v in values])
        for val, ref in zip(m.initial_values(True), values):
            self.assertEqual(val, ref)

        # Set with mix, including strings and expressions
        values[2] += 10
        mixed = list(values)
        mixed[2] = myokit.Number(mixed[2])
        mixed[3] = str(values[3])
        m.set_initial_values(mixed)
        for val, ref in zip(m.initial_values(True), values):
            self.assertEqual(val, ref)

        # Set with dict of floats
        values = m.initial_values(as_floats=True)
        s = dict(zip(states, values))
        m.set_initial_values(s)
        for val, ref in zip(m.initial_values(True), values):
            self.assertEqual(val, ref)

        # Set with one big string
        values[0] += 5
        s = dict(zip(states, values))
        s = '\n'.join([str(a) + '=' + str(b) for a, b in s.items()])
        m.set_initial_values(s)
        for val, ref in zip(m.initial_values(True), values):
            self.assertEqual(val, ref)

        # Test that errors are detected in validation
        values[0] = myokit.Name(m.get('membrane.V'))
        m.set_initial_values(values)
        self.assertRaisesRegex(myokit.IntegrityError, 'not const', m.validate)
        values[0] = 'sqrt(3)'
        m.set_initial_values(values)
        m.validate()

        # Deprecated alias
        values = m.initial_values(as_floats=True)
        values[0] += 10
        with WarningCollector() as w:
            m.set_state(values)
        self.assertIn('deprecated', w.text())
        for val, ref in zip(m.initial_values(True), values):
            self.assertEqual(val, ref)

        # Wrong number of elements
        self.assertRaisesRegex(
            ValueError, 'Wrong number of initial values',
            m.set_initial_values, [1, 2, 3])

    def test_show_evaluation_of(self):
        # Test :meth:`Model.show_evaluation_of(variable)`.
        # Depends mostly on `references()`, and `code()` methods.

        m = myokit.load_model('example')

        # Test for literal
        e = m.show_evaluation_of('cell.Na_o')
        self.assertIn('cell.Na_o = ', e)
        self.assertIn('Literal constant', e)
        self.assertEqual(len(e.splitlines()), 6)

        # Test for calculated constant
        e = m.show_evaluation_of('ina.ENa')
        self.assertIn('ina.ENa = ', e)
        self.assertIn('Calculated constant', e)
        self.assertEqual(len(e.splitlines()), 12)

        # Test for intermediary variable
        e = m.show_evaluation_of('ina.INa')
        self.assertIn('ina.INa = ', e)
        self.assertIn('Intermediary variable', e)
        self.assertEqual(len(e.splitlines()), 15)

        # Test for state variable (with nested variables)
        e = m.show_evaluation_of('ina.m')
        self.assertIn('ina.m = ', e)
        self.assertIn('State variable', e)
        self.assertEqual(len(e.splitlines()), 15)

        # Test with guessing of similar
        e = m.show_evaluation_of('ina.Na_o')
        self.assertIn('not found', e)
        self.assertIn('cell.Na_o = ', e)
        self.assertIn('Literal constant', e)
        self.assertEqual(len(e.splitlines()), 7)

        # Test with expressions in initial values
        m.get('ina.m').set_initial_value('1 / sqrt(4)')
        e = m.show_evaluation_of('ina.m')
        self.assertIn('Initial value = 1 / sqrt(4)', e)
        self.assertEqual(len(e.splitlines()), 16)
        m.get('ina.m').set_initial_value('1 / sqrt(ina.gNa)')
        e = m.show_evaluation_of('ina.m')
        self.assertIn('Initial value = 1 / sqrt(ina.gNa)', e)
        self.assertIn('              = 2.5000000', e)

        # Test with nothing similar
        m = myokit.Model()
        self.assertRaises(Exception, m.show_evaluation_of, 'Hello')

    def test_show_expressions_for(self):
        # Test :meth:`Model.show_expressions_for(variable)`.

        m = myokit.load_model('example')
        e = m.show_expressions_for(m.get('ina.INa'))
        self.assertIn('ina.INa is a function of', e)
        self.assertEqual(len(e.splitlines()), 22)

    def test_show_line_of(self):
        # Test :meth:`Model.show_line_of(variable)`.

        # Check string with info
        m = myokit.load_model('example')
        v = m.get('ina.INa')
        e = m.show_line_of(v)
        self.assertIn('Defined on line 90', e)
        self.assertIn('Intermediary variable', e)
        self.assertEqual(len(e.splitlines()), 4)

        # Check with freshly made model
        m2 = m.clone()
        v2 = m2.get('ina.INa')
        e = m2.show_line_of(v2)
        self.assertNotIn('Defined on line', e)
        self.assertIn('Intermediary variable', e)
        self.assertEqual(len(e.splitlines()), 3)

        # 'raw' version
        self.assertEqual(m.show_line_of(v, raw=True), 90)
        self.assertIsNone(m2.show_line_of(v2, raw=True))

    def test_str(self):
        # Test conversion to string

        m = myokit.Model()
        self.assertEqual(str(m), '<Unnamed Model>')
        self.assertEqual(repr(m), '<Unnamed Model>')

        m.set_name('bert')
        self.assertEqual(str(m), '<Model(bert)>')
        self.assertEqual(repr(m), '<Model(bert)>')

    def test_suggest(self):
        # Test :meth:`Model.suggest(variable_name)`.

        m = myokit.Model()
        c1 = m.add_component('c1')
        t = c1.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        c2 = m.add_component('c2')
        v = c2.add_variable('v')
        v.set_rhs('3')

        # Test with correct name
        found, suggested, msg = m.suggest_variable('c1.time')
        self.assertEqual(found, t)
        self.assertIsNone(suggested)
        self.assertIsNone(msg)

        # Test with wrong name
        found, suggested, msg = m.suggest_variable('c1.tim')
        self.assertIsNone(found)
        self.assertEqual(suggested, t)
        self.assertIn('Unknown', msg)

        # Test with case mismatch
        found, suggested, msg = m.suggest_variable('c1.timE')
        self.assertIsNone(found)
        self.assertEqual(suggested, t)
        self.assertIn('Case mismatch', msg)

        # Test without component
        found, suggested, msg = m.suggest_variable('time')
        self.assertIsNone(found)
        self.assertEqual(suggested, t)
        self.assertIn('No component', msg)

    def test_unique_names_1(self):
        # Test Model.create_unique_names().

        # Heavily disputed variable names
        m = myokit.Model()
        a = m.add_component('a')
        ax = a.add_variable('x')
        b = m.add_component('b')
        bx = b.add_variable('x')
        x = m.add_component('x')
        xx = x.add_variable('x')
        m.create_unique_names()
        self.assertEqual(a.uname(), 'a')
        self.assertEqual(ax.uname(), 'a_x')
        self.assertEqual(b.uname(), 'b')
        self.assertEqual(bx.uname(), 'b_x')
        self.assertEqual(x.uname(), 'x_1')
        self.assertEqual(xx.uname(), 'x_x')

        # Disputed variable name --> Generated name already exists
        m = myokit.Model()
        a = m.add_component('a')
        ax = a.add_variable('x')
        abx = a.add_variable('b_x')
        aax = a.add_variable('a_x')
        ax11 = a.add_variable('x_1_1')
        b = m.add_component('b')
        bx = b.add_variable('x')
        bx11 = b.add_variable('x_1_1')
        m.create_unique_names()
        self.assertEqual(a.uname(), 'a')
        self.assertEqual(ax.uname(), 'a_x_1')
        self.assertEqual(abx.uname(), 'b_x')
        self.assertEqual(aax.uname(), 'a_x')
        self.assertEqual(ax11.uname(), 'a_x_1_1')
        self.assertEqual(b.uname(), 'b')
        self.assertEqual(bx.uname(), 'b_x_1')
        self.assertEqual(bx11.uname(), 'b_x_1_1')

        # Disputed component name
        m = myokit.Model()
        a = m.add_component('a')
        a.add_variable('x')
        m.add_component('x')
        m.create_unique_names()
        self.assertEqual(m.get('a').uname(), 'a')
        self.assertEqual(m.get('a.x').uname(), 'a_x')
        self.assertEqual(m.get('x').uname(), 'x_1')

        # Disputed component name --> Generated name already exists
        m = myokit.Model()
        a = m.add_component('a')
        a.add_variable('x')
        m.add_component('x')
        m.add_component('x_1')
        m.create_unique_names()
        self.assertEqual(m.get('a').uname(), 'a')
        self.assertEqual(m.get('a.x').uname(), 'a_x')
        self.assertEqual(m.get('x').uname(), 'x_2')
        self.assertEqual(m.get('x_1').uname(), 'x_1')

    def test_unique_names_2(self):
        # Test reserving of unique name prefixes

        m = myokit.Model()
        a = m.add_component('paddington')
        x = a.add_variable('v_x')
        y = a.add_variable('bear')
        b = m.add_component('yogi')
        z = b.add_variable('bear')

        m.create_unique_names()
        self.assertEqual(x.uname(), 'v_x')
        self.assertEqual(y.uname(), 'paddington_bear')
        self.assertEqual(z.uname(), 'yogi_bear')

        # Don't allow v_ prefix
        m.reserve_unique_name_prefix('v_', 'var_')
        m.create_unique_names()
        self.assertEqual(x.uname(), 'var_v_x')
        self.assertEqual(y.uname(), 'paddington_bear')
        self.assertEqual(z.uname(), 'yogi_bear')

        # Don't allow pad prefix
        m.reserve_unique_name_prefix('pad', 'marmelade_')
        m.create_unique_names()
        self.assertEqual(x.uname(), 'var_v_x')
        self.assertEqual(y.uname(), 'marmelade_paddington_bear')
        self.assertEqual(z.uname(), 'yogi_bear')
        self.assertEqual(a.uname(), 'marmelade_paddington')

        # Test bad calls
        self.assertRaisesRegex(
            ValueError, 'prefix cannot be empty',
            m.reserve_unique_name_prefix, '', 'x')
        self.assertRaisesRegex(
            ValueError, 'prepend cannot be empty',
            m.reserve_unique_name_prefix, 'x', '')
        self.assertRaisesRegex(
            ValueError, 'prepend cannot start with prefix',
            m.reserve_unique_name_prefix, 'x', 'x')

    def test_unused_and_cycles(self):
        # Test unused variable and cycle detection.

        m = myokit.Model('LotkaVolterra')
        c0 = m.add_component('c0')
        t = c0.add_variable('time')
        t.set_rhs(myokit.Number(0))
        t.set_binding('time')
        c1 = m.add_component('c1')
        m.add_component('c2')
        c1_a = c1.add_variable('a')
        c1_b = c1.add_variable('b')
        c1_a.promote(1.0)
        c1_a.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Number(0.5)))
        c1_b.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Number(1.0)))
        # b is unused, test if found
        m.validate()
        w = m.warnings()
        self.assertEqual(len(w), 1)
        self.assertEqual(type(w[0]), myokit.UnusedVariableError)
        # b is used by c, c is unused, test if found
        c1_c = c1.add_variable('c')
        c1_c.set_rhs(myokit.Name(c1_b))
        m.validate()
        w = m.warnings()
        self.assertEqual(len(w), 2)
        self.assertEqual(type(w[0]), myokit.UnusedVariableError)
        self.assertEqual(type(w[1]), myokit.UnusedVariableError)
        # Test 1:1 cycle
        c1_b.set_rhs(myokit.Name(c1_b))
        self.assertRaises(myokit.CyclicalDependencyError, m.validate)
        # Test longer cycles
        c1_b.set_rhs(myokit.Multiply(myokit.Number(10), myokit.Name(c1_c)))
        self.assertRaises(myokit.CyclicalDependencyError, m.validate)
        # Reset
        c1_b.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Number(1.0)))
        m.validate()
        # Test cycle involving state variable
        c1_a.set_rhs(myokit.Name(c1_b))
        m.validate()
        c1_b.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Name(c1_b)))
        self.assertRaises(myokit.CyclicalDependencyError, m.validate)
        c1_b.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Name(c1_c)))
        c1_c.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Number(3)))
        m.validate()
        w = m.warnings()
        self.assertEqual(len(w), 0)
        c1_c.set_rhs(myokit.Multiply(myokit.Name(c1_a), myokit.Name(c1_b)))
        self.assertRaises(myokit.CyclicalDependencyError, m.validate)

    def test_validate_and_remove_unused_variables(self):
        # Test :class:`Model.validate` with ``remove_unused_variables=True``.

        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        x = c.add_variable('x')
        y = c.add_variable('y')
        z = c.add_variable('z')
        z1 = z.add_variable('z1')
        x.set_rhs('(10 - x) / y')
        x.promote(0)
        y.set_rhs(1)
        z.set_rhs('2 + z1')
        z1.set_rhs(3)

        # Two unused variables: z and z1
        m.validate()
        self.assertEqual(len(m.warnings()), 2)

        # Remove unused variables
        m.validate(remove_unused_variables=True)
        self.assertEqual(len(m.warnings()), 2)  # 2 removal warnings
        m.validate()
        self.assertEqual(len(m.warnings()), 0)  # issue fixed!

    def test_value(self):
        # Test :meth:`Model.value()`.

        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_binding('time')
        t.set_rhs(1000)
        self.assertEqual(m.value('c.t'), 1000)

    def test_warnings(self):
        # Test Model.has_warnings(), model.warnings() and
        # Model.format_warnings().

        # Test model without warnings
        m = myokit.Model()
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_binding('time')
        t.set_rhs(0)
        v = c.add_variable('v')
        v.set_rhs('3 - v')
        v.promote(0.1)
        m.validate()

        self.assertFalse(m.has_warnings())
        self.assertIn('0 validation warning', m.format_warnings())
        self.assertEqual(m.warnings(), [])

        # Test model with warnings
        v.validate()
        v.demote()
        v.validate()
        v.set_rhs(3)
        m.validate()

        self.assertTrue(m.has_warnings())
        self.assertIn('1 validation warning', m.format_warnings())
        self.assertIn('Unused variable', str(m.warnings()[0]))


if __name__ == '__main__':
    unittest.main()




#
# Functions related to floating point numbers. Imported by default.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import sys

import myokit

# Store ref to built-in `round` and `str` functions
_round = round
_str = str


def close(a, b, reltol=1e-9, abstol=1e-9):
    """
    Test whether two numbers are close enough to be considered equal.

    Differs from :meth:`myokit.float.eq` in that it tries to answer the
    question "are two numbers resulting from various calculations close enough
    to be considered equal". Whereas :meth`myokit.float.eq` aims to deal with
    umbers that are numerically indistinguishable but still have a slightly
    different floating point representation.
    """
    # Note the initial == check handles infinity
    return a == b or abs(a - b) < max(reltol * max(abs(a), abs(b)), abstol)


def cround(x, reltol=1e-9, abstol=1e-9):
    """
    Checks if a float ``x`` is close to an integer with :meth:`close()`, and if
    so, returns that integer.
    """
    ix = int(_round(x))
    return ix if close(x, ix, reltol, abstol) else x


def eq(a, b):
    """
    Checks if floating point numbers ``a`` and ``b`` are equal, or so close to
    each other that the difference could be a single rounding error.
    """
    # Note the initial == check handles infinity
    return a == b or abs(a - b) < max(abs(a), abs(b)) * sys.float_info.epsilon


def geq(a, b):
    """
    Checks if ``a >= b``, but using :meth:`myokit.float.eq` instead of ``==``.
    """
    return a >= b or abs(a - b) < max(abs(a), abs(b)) * sys.float_info.epsilon


def round(x):
    """
    Checks if a float ``x`` is within a single rounding error of an integer,
    and if so, returns that integer.
    """
    ix = int(_round(x))
    return ix if eq(x, ix) else x


def str(number, full=False, precision=myokit.DOUBLE_PRECISION):
    """
    Converts the given number to a string, using a full-precision
    representation if needed.

    Arguments:

    ``full``
        Set to ``True`` to force full precision output.
    ``precision``
        Set to ``myokit.SINGLE_PRECISION`` to assume single-precision numbers
        when formatting with full precision.

    """
    # Force full precision output
    if full:
        if precision == myokit.SINGLE_PRECISION:
            return myokit.SFSINGLE.format(float(number))
        else:
            return myokit.SFDOUBLE.format(float(number))

    # Pass through strings
    if isinstance(number, _str):
        return number

    # Handle myokit.Numbers
    if isinstance(number, myokit.Number):
        number = number.eval()

    # For most numbers, allow python to format the float
    s = _str(number)
    if len(s) < 10:
        return s

    # But if the number is given with lots of decimals, use the representation
    # with enough digits to prevent loss of information
    if precision == myokit.SINGLE_PRECISION:
        return myokit.SFSINGLE.format(float(number))
    else:
        return myokit.SFDOUBLE.format(float(number))




#
# Exports to a CUDA kernel
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit.formats


class CudaKernelExporter(myokit.formats.TemplatedRunnableExporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` creates an unoptimised
    CUDA cell kernel that calculates a forward Euler step for a single cell.

    Only a kernel file is created, no driver class is included and no support
    for protocol export is provided.

    A value must be bound to ``diffusion_current`` which represents the
    current flowing from cell to cell. This is defined as positive when the
    cell is acting as a source, negative when it acts like a sink. In other
    words, it is defined as an outward current.

    Please keep in mind that CellML and other downloaded formats are typically
    not directly suitable for GPU simulation. Specifically, when simulating
    on single-precision devices a lot of divide-by-zero errors might crop up
    that remain hidden when using double precision single cell simulations on
    the CPU.
    """

    def _dir(self, root):
        return os.path.join(root, 'cuda', 'template')

    def _dict(self):
        return {'kernel.cu': 'kernel.cu'}

    def _vars(self, model, protocol):
        return {
            'model': model,
            'use_rl': False,
        }


class CudaKernelRLExporter(myokit.formats.TemplatedRunnableExporter):
    """
    Like :class:`CudaKernelExporter` but uses a Rush-Larsen update step where
    applicable.

    For this exporter to work, the model's membrane potential variable must be
    labelled as ``membrane_potential``, and a value must be bound to
    ``diffusion_current`` (see :class:`CudaKernelExporter` for details).
    """

    def _dir(self, root):
        return os.path.join(root, 'cuda', 'template')

    def _dict(self):
        return {'kernel.cu': 'kernel.cu'}

    def _vars(self, model, protocol):
        return {
            'model': model,
            'use_rl': True,
        }



#
# Tools for working with Markov models of ion channels.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections

import numpy as np

import myokit


class LinearModel:
    """
    Represents a linear Markov model of an ion channel extracted from a
    :class:`myokit.Model`.

    The class assumes the markov model can be written as::

        dot(x) = A(V,p) * x
        I = B(V,p) * x

    where ``V`` is the membrane potential, ``p`` is a set of parameters and
    ``A`` and ``B`` are the matrices that relate the state ``x`` to its
    derivative ``dot(x)`` and a current ``I``.

    ``A`` and ``B`` can contain non-linear functions, but should be simple
    scalar matrices when evaluated for a fixed ``V`` and ``p``. For example,
    the current equation ``I = g * O * (V - E)`` would have ``p = [g, E]``, so
    that for fixed ``p`` and ``V`` this would resolve to
    ``I = (g * (V - E)) * O``, such that ``g * (V - E)`` is a constant that can
    be included in ``B``.

    The model variables to treat as parameter are specified by the user when
    the model is created. Any other variables, for example state variables such
    as intercellular calcium or constants such as temperature, are fixed when
    the markov model is created and can no longer be changed. Initial values
    written as expressions are evaluated when the model is made.

    To create a :class:`Markov`, pass in a :class:`myokit.Model` and select a
    list of states. All other states will be fixed at their current value and
    an attempt will be made to write the remaining state equations as linear
    combinations of the states. If this is not possible, a :class:`ValueError`
    is raised. The membrane potential must be indicated using the label
    ``membrane_potential`` or by passing it in as ``vm``.

    The current variable is optional, if no current is specified by the user
    the relation ``I = B * x`` is dropped and no ``B`` is calculated.

    Example::

        import myokit
        import myokit.lib.markov as markov

        # Load a model from disk
        model = myokit.load_model('some-model.mmt')

        # Select the relevant states and parameters
        states = [
            'ina.C1',
            'ina.C2',
            'ina.O',
            ...
            ]
        parameters = [
            'ina.p1',
            'ina.p2',
            ...
            ]
        current = 'ina.INa'

        # Extract a markov model
        mm = markov.LinearModel(model, states, parameters, current)

        # Get the matrices A and B such that dot(x) = A * x and I = B * x
        # where ``x`` is the state vector and ``I`` is the current.
        A, B = mm.matrices(membrane_potential=-40)
        print(A)

    Alternatively, a LinearModel can be constructed from a single component
    using the method :meth:`from_component() <LinearModel.from_component()>`::

        import myokit
        import myokit.lib.markov as markov

        # Load a model from disk
        model = myokit.load_model('some-model.mmt')

        # Extract a markov model
        mm = markov.LinearModel.from_component(model.get('ina'))

    Arguments:

    ``model``
        The model to work with.
    ``states``
        An ordered list of state variables (or state variable names) from
        ``model``. All remaining state variables will be frozen in place. Each
        state's derivative must be a linear combination of the other states.
    ``parameters``
        A list of parameters to maintain in their symbolic form.
    ``current``
        The markov model's current variable. The current must be a linear
        combination of the states (for example ``g * (V - E) * (O1 + O2)``)
        where ``O1`` and ``O2`` are states. If no current variable is specified
        ``None`` can be used instead.
    ``vm``
        The variable indicating membrane potential. If set to ``None``
        (default) the method will search for a variable with the label
        ``membrane_potential``.

    """
    # NOTE: A LinearModel must be immutable!
    # This ensures that the simulations don't have to clone it (which would be
    # costly and difficult).
    # A LinearModel has a method "matrices()" which can evaluate A and B
    # for its default values or newly passed in values, but it never updates
    # its internal state in any way!
    def __init__(self, model, states, parameters=None, current=None, vm=None):
        super().__init__()

        # Get a clone of the model, with all markov models written in full ODE
        # form.
        self._model = convert_markov_models_to_full_ode_form(model)
        del model

        #
        # Check input
        #

        # Check and collect state variables
        self._states = []
        for state in states:
            if isinstance(state, myokit.Variable):
                state = state.qname()
            try:
                state = self._model.get(str(state), myokit.Variable)
            except KeyError:
                raise LinearModelError('Unknown state: <' + str(state) + '>.')
            if not state.is_state():
                raise LinearModelError(
                    'Variable <' + state.qname() + '> is not a state.')
            if state in self._states:
                raise LinearModelError(
                    'State <' + state.qname() + '> was added twice.')
            self._states.append(state)
        del states

        # Check and collect parameter variables
        unique = set()
        self._parameters = []
        if parameters is None:
            parameters = []
        for parameter in parameters:
            if isinstance(parameter, myokit.Variable):
                parameter = parameter.qname()
            try:
                parameter = self._model.get(parameter, myokit.Variable)
            except KeyError:
                raise LinearModelError(
                    'Unknown parameter: <' + str(parameter) + '>.')
            if not parameter.is_literal():
                raise LinearModelError(
                    'Unsuitable parameter: <' + str(parameter) + '>.')
            if parameter in unique:
                raise LinearModelError(
                    'Parameter listed twice: <' + str(parameter) + '>.')
            unique.add(parameter)
            self._parameters.append(parameter)
        del unique
        del parameters

        # Check current variable
        if current is not None:
            if isinstance(current, myokit.Variable):
                current = current.qname()
            current = self._model.get(current, myokit.Variable)
            if current.is_state():
                raise LinearModelError('Current variable can not be a state.')
        self._current = current
        del current

        # Check membrane potential variable
        if vm is None:
            vm = self._model.label('membrane_potential')
        if vm is None:
            raise LinearModelError(
                'A membrane potential must be specified as `vm` or using the'
                ' label `membrane_potential`.')
        if isinstance(vm, myokit.Variable):
            vm = vm.qname()
        self._membrane_potential = self._model.get(vm)
        if self._membrane_potential in self._parameters:
            raise LinearModelError(
                'Membrane potential should not be included in the list of'
                ' parameters.')
        if self._membrane_potential in self._states:
            raise LinearModelError(
                'The membrane potential should not be included in the list of'
                ' states.')
        if self._membrane_potential == self._current:
            raise LinearModelError(
                'The membrane potential should not be the current variable.')
        del vm

        # Get values of all states
        # Note: Do this _before_ changing the model!
        s = self._model.initial_values(True)
        self._default_state = np.array(
            [v.initial_value(True) for v in self._states])

        #
        # Demote unnecessary states, remove bindings and validate model.
        #

        # Freeze remaining, non-markov-model states
        for k, state in enumerate(self._model.states()):
            if state not in self._states:
                state.demote()
                state.set_rhs(s[k])
        del s

        # Unbind everything except time
        for label, var in self._model.bindings():
            if label != 'time':
                var.set_binding(None)

        # Check if current variable depends on selected states
        # (At this point, anything not dependent on the states is a constant)
        if self._current is not None and self._current.is_constant():
            raise LinearModelError(
                'Current must be a function of the markov model\'s state'
                ' variables.')

        # Validate modified model
        self._model.validate()

        #
        # Create functions:
        #   matrix_function(vm, p1, p2, ...   ) --> A, B
        #       where dot(x) = Ax and I = Bx
        #   rate_list_function(vm, p1, p2, ...) --> R
        #       where R contains tuples (i, j, rij)
        #

        # Create a list of inputs to the functions
        self._inputs = self._parameters + [self._membrane_potential]

        # Get the default values for all inputs
        self._default_inputs = np.array([v.eval() for v in self._inputs])

        # Create functions
        self._matrix_function = None
        self._rate_list_function = None
        self._generate_functions()

        #
        # Partial validation
        #
        # Check if dependencies are bidirectional
        for s in self._states:
            for d in s.refs_to(state_refs=True):
                if s not in d.refs_to(state_refs=True):
                    raise LinearModelError(
                        'State <' + s.qname() + '> depends on <' + d.qname()
                        + '> but not vice versa.')

        # Check the sum of all states is 1
        tolerance = 1e-8
        x = np.sum(self._default_state)
        if np.abs(x - 1) > tolerance:
            raise LinearModelError(
                'The sum of states is not equal to 1: ' + str(x))

        # Check the sum of all derivatives per column
        A, B = self.matrices()
        for k, x in enumerate(np.sum(A, axis=0)):
            if abs(x) > tolerance:
                raise LinearModelError(
                    'Derivatives in column ' + str(1 + k) + ' sum to non-zero'
                    ' value: ' + str(x) + '.')

    def _generate_functions(self):
        """
        Creates a function that takes parameters as input and returns matrices
        A and B.

        (This method is called only once, by the constructor, but it's
        complicated enough to warrant its own method...)
        """
        # Create mapping from states to index
        state_indices = {}
        for k, state in enumerate(self._states):
            state_indices[state] = k

        # Get expressions for state & current variables, but with all
        # references to variables (except states and parameters) replaced by
        # inlined expressions.
        expressions = []
        for v in self._states:
            expressions.append(v.rhs().clone(expand=True, retain=self._inputs))
        if self._current is not None:
            current_expression = self._current.rhs().clone(
                expand=True, retain=self._inputs)

        # Create parametrisable matrices to evaluate the state & current
        # This checks that each state's RHS can be written as a linear
        # combination of the states, and gathers the corresponding multipliers.

        # Matrix of linear factors
        n = len(self._states)
        A = [[myokit.Number(0) for j in range(n)] for i in range(n)]

        # List of transitions
        T = set()

        # Populate A and T
        for row, e in enumerate(expressions):

            # Check if this expression is a linear combination of the states
            try:
                factors = _linear_combination(e, self._states)
            except ValueError:
                raise LinearModelError(
                    'Unable to write expression as linear combination of'
                    ' states: ' + str(e) + '.')

            # Scan factors
            for col, state in enumerate(self._states):
                factor = factors[col]
                if factor is not None:
                    # Add factor to transition matrix
                    A[row][col] = factor

                    # Store transition in transition list
                    if row != col:
                        T.add((col, row))   # A is mirrored

        # Create a parametrisable matrix for the current
        B = [myokit.Number(0) for i in range(n)]
        if self._current is not None:
            try:
                factors = _linear_combination(current_expression, self._states)
            except ValueError:
                raise LinearModelError(
                    'Unable to write expression as linear combination of'
                    ' states: ' + str(e) + '.')

            for col, state in enumerate(self._states):
                factor = factors[col]
                if factor is not None:
                    B[col] = factor

        # Create list of transition rates and associated equations
        T = list(T)
        T.sort()
        R = []
        for i in range(len(A)):
            for j in range(len(A)):
                if (i, j) in T:
                    R.append((i, j, A[j][i]))   # A is mirrored
        del T

        #
        # Create function to create parametrisable matrices
        #
        self._model.reserve_unique_names('A', 'B', 'n', 'numpy')
        writer = myokit.numpy_writer()
        w = writer.ex
        head = 'def matrix_function('
        head += ','.join([w(p.lhs()) for p in self._inputs])
        head += '):'
        body = []
        body.append('A = numpy.zeros((n, n))')
        zero = myokit.Number(0)
        for i, row in enumerate(A):
            for j, e in enumerate(row):
                if e != zero:
                    body.append('A[' + str(i) + ',' + str(j) + '] = ' + w(e))
        body.append('B = numpy.zeros(n)')
        for j, e in enumerate(B):
            if e != zero:
                body.append('B[' + str(j) + '] = ' + w(e))
        body.append('return A, B')
        code = head + '\n' + '\n'.join(['    ' + line for line in body])
        globl = {'numpy': np, 'n': n}
        local = {}

        exec(code, globl, local)
        self._matrix_function = local['matrix_function']

        #
        # Create function to return list of transition rates
        #
        self._model.reserve_unique_names('R', 'n', 'numpy')
        head = 'def rate_list_function('
        head += ','.join([w(p.lhs()) for p in self._inputs])
        head += '):'
        body = []
        body.append('R = []')
        for i, j, e in R:
            body.append(
                'R.append((' + str(i) + ',' + str(j) + ',' + w(e) + '))')
        body.append('return R')
        code = head + '\n' + '\n'.join(['    ' + line for line in body])
        globl = {'numpy': np}
        local = {}
        exec(code, globl, local)
        self._rate_list_function = local['rate_list_function']

    def current(self):
        """
        Returns the name of the current variable used by this model, or None if
        no current variable was specified.
        """
        return self._current

    def default_membrane_potential(self):
        """
        Returns this markov model's default membrane potential value.
        """
        return self._default_inputs[-1]

    def default_parameters(self):
        """
        Returns this markov model's default parameter values
        """
        return list(self._default_inputs[:-1])

    def default_state(self):
        """
        Returns this markov model's default state values.
        """
        return list(self._default_state)

    @staticmethod
    def from_component(
            component, states=None, parameters=None, current=None, vm=None):
        """
        Creates a Markov model from a component, using the following rules:

          1. Every state in the component is a state in the Markov model
          2. Every unnested constant in the component is a parameter
          3. The component should contain exactly one unnested intermediary
             variable whose value depends on the model states, this will be
             used as the current variable.
          4. The model contains a variable labeled "membrane_potential".

        Any of the automatically set variables can be overridden using the
        keyword arguments ``states``, ``parameters``, ``current`` and ``vm``.

        The parameters, if determined automatically, will be specified in
        alphabetical order (using a natural sort).
        """
        model = component.model()

        # Get or check states
        if states is None:

            # Get state variables
            states = [x for x in component.variables(state=True)]

            # Sort by state index
            states.sort(key=lambda x: x.index())

        else:

            # Make sure states are variables. This is required to automatically
            # find a current variable.
            states_in = states
            states = []
            for state in states_in:
                if isinstance(state, myokit.Variable):
                    state = state.qname()
                states.append(model.get(state))

        # Get parameters
        if parameters is None:

            # Get parameters
            parameters = [
                x for x in component.variables(const=True) if x.is_literal()]
            # Sort by qname, using natural sort
            parameters.sort(
                key=lambda x: myokit.tools.natural_sort_key(x.qname()))

        # Get current
        if current is None:
            currents = []
            for x in component.variables(inter=True):
                for y in x.refs_to(state_refs=True):
                    if y in states:
                        # Found a candidate!
                        currents.append(x)
                        break
            if len(currents) > 1:
                raise LinearModelError(
                    'The given component has more than one variable that could'
                    ' be a current: '
                    + ', '.join(['<' + x.qname() + '>' for x in currents])
                    + '.')
            try:
                current = currents[0]
            except IndexError:
                pass

        # Get membrane potential
        if vm is None:
            vm = model.label('membrane_potential')
            if vm is None:
                raise LinearModelError(
                    'The model must define a variable labeled as'
                    ' "membrane_potential".')

        # Create and return LinearModel
        return LinearModel(model, states, parameters, current, vm)

    def matrices(self, membrane_potential=None, parameters=None):
        """
        For a given value of the ``membrane_potential`` and a list of values
        for the ``parameters``, this method calculates and returns the matrices
        ``A`` and ``B`` such that::

            dot(x) = A * x
            I = B * x

        where ``x`` is the state vector and ``I`` is the current.

        Arguments:

        ``membrane_potential``
            The value to use for the membrane potential, or ``None`` to use the
            value from the original :class:`myokit.Model`.
        ``parameters``
            The values to use for the parameters, given in the order they were
            originally specified in (if the model was created using
            :meth:`from_component()`, this will be alphabetical order).

        """
        inputs = list(self._default_inputs)
        if membrane_potential is not None:
            inputs[-1] = float(membrane_potential)
        if parameters is not None:
            if len(parameters) != len(self._parameters):
                raise ValueError(
                    'Illegal parameter vector size: '
                    + str(len(self._parameters)) + ' required, '
                    + str(len(parameters)) + ' provided.')
            inputs[:-1] = [float(x) for x in parameters]
        return self._matrix_function(*inputs)

    def membrane_potential(self):
        """
        Returns the name of the membrane potential variable used by this model.
        """
        return self._membrane_potential.qname()

    def parameters(self):
        """
        Returns the names of the parameter variables used by this model.
        """
        return [v.qname() for v in self._parameters]

    def rates(self, membrane_potential=None, parameters=None):
        """
        For a given value of the ``membrane_potential`` and a list of values
        for the ``parameters``, this method calculates and returns an ordered
        list of tuples ``(i, j, rij)`` such that ``rij`` is a non-zero
        transition rate from the ``i``-th state to the ``j``-th state.

        Arguments:

        ``membrane_potential``
            The value to use for the membrane potential, or ``None`` to use the
            value from the original :class:`myokit.Model`.
        ``parameters``
            The values to use for the parameters, given in the order they were
            originally specified in (if the model was created using
            :meth:`from_component()`, this will be alphabetical order).

        """
        inputs = list(self._default_inputs)
        if membrane_potential is not None:
            inputs[-1] = float(membrane_potential)
        if parameters is not None:
            if len(parameters) != len(self._parameters):
                raise ValueError(
                    'Illegal parameter vector size: '
                    + str(len(self._parameters)) + ' required, '
                    + str(len(parameters)) + ' provided.')
            inputs[:-1] = [float(x) for x in parameters]
        return self._rate_list_function(*inputs)

    def states(self):
        """
        Returns the names of the state variables used by this model.
        """
        return [v.qname() for v in self._states]

    def steady_state(self, membrane_potential=None, parameters=None):
        """
        Analytically determines a steady state solution for this Markov model.

        ``membrane_potential``
            The value to use for the membrane potential, or ``None`` to use the
            value from the original :class:`myokit.Model`.
        ``parameters``
            The values to use for the parameters, given in the order they were
            originally specified in (if the model was created using
            :meth:`from_component()`, this will be alphabetical order).

        """
        # Calculate Jacobian and derivatives
        A, _ = self.matrices(membrane_potential, parameters)

        # Set up reduced system with full rank: dot(x) = Ay + B
        B = A[:-1, -1:]
        A = A[:-1, :-1] - B

        # Check eigenvalues
        if np.max(np.linalg.eigvals(A) >= 0):
            raise LinearModelError(
                'System has positive eigenvalues: won\'t converge to steady'
                ' state!')

        # Solve system Ax + B = 0 --> Ax = -B
        x = np.linalg.solve(A, -B)

        # Recreate full state vector and return
        x = np.array(x).reshape((len(x),))
        x = np.concatenate((x, [1 - np.sum(x)]))
        return x


class AnalyticalSimulation:
    """
    Analytically evaluates a :class:`LinearModel`'s state over a given set of
    points in time.

    Solutions are calculated for the "law of large numbers" case, i.e. without
    stochastic behavior. The solution algorithm is based on eigenvalue
    decomposition.

    Each simulation object maintains an internal state consisting of

    * The current simulation time
    * The current state
    * The default state

    When a simulation is created, the simulation time is set to zero and both
    the current and default state are initialized using the ``LinearModel``.
    After each call to :meth:`run()` the time and current state are updated,
    so that each successive call to run continues where the previous simulation
    left off.

    A :class:`protocol <myokit.Protocol>` can be used to set the membrane
    potential during the simulation, or the membrane potential can be adjusted
    manually between runs.

    Example::

        import myokit
        import myokit.lib.markov as markov

        # Create a linear markov model
        m = myokit.load_model('clancy-1999.mmt')
        m = markov.LinearModel.from_component(m.get('ina'))

        # Create an analytical simulation object
        s = markov.AnalyticalSimulation(m)

        # Run a simulation
        s.set_membrane_potential(-30)
        d = s.run(10)

        # Show the results
        import matplotlib.pyplot as plt
        plt.figure()
        plt.subplot(211)
        for state in m.states():
            plt.plot(d.time(), d[state], label=state)
        plt.legend(loc='center right')
        plt.subplot(212)
        plt.plot(d.time(), d[m.current()])
        plt.show()

    """
    def __init__(self, model, protocol=None):
        super().__init__()
        # Check model
        if not isinstance(model, LinearModel):
            raise ValueError('First parameter must be a `LinearModel`.')
        self._model = model

        # Check protocol
        if protocol is None:
            self._protocol = None
        elif not isinstance(protocol, myokit.Protocol):
            raise ValueError('Protocol must be a myokit.Protocol object')
        else:
            self._protocol = protocol.clone()

        # Time variable
        self._time = 0

        # Check if we have a current variable
        self._has_current = self._model.current() is not None

        # Set state
        self._state = np.array(
            self._model.default_state(), copy=True, dtype=float)

        # Set default state
        self._default_state = np.array(self._state, copy=True)

        # Get membrane potential
        self._membrane_potential = self._model.default_membrane_potential()

        # Get parameters
        self._parameters = np.array(
            self._model.default_parameters(), copy=True, dtype=float)

        # Mapping from parameter names to index in parameter array
        self._parameter_map = {}
        for i, p in enumerate(self._model.parameters()):
            self._parameter_map[p] = i

        # Cached matrices and partial solution (eigenvalue decomposition etc.)
        # Both stored per voltage, but will become invalidated if parameters
        # change
        self._cached_matrices = {}
        self._cached_solution = {}

        # If protocol was given, create pacing system, update vm
        self._pacing = None
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

        # Keys for logging
        self._time_key = self._model._model.time().qname()
        self._vm_key = self._model._membrane_potential
        self._log_keys = [self._time_key, self._vm_key] + self._model.states()
        if self._has_current:
            self._log_keys.append(self._model.current())

    def current(self, state):
        """
        Calculates the current for a given state.
        """
        if not self._has_current:
            raise Exception(
                'The used model did not specify a current variable.')
        A, B = self._matrices()
        return B.dot(state)

    def default_state(self):
        """
        Returns the default state used by this simulation.
        """
        return list(self._default_state)

    def _matrices(self):
        """
        Returns the (cached or re-generated) matrices.

        Returns ``(A, B)`` if this simulation's model has a current variable,
        or just ``B`` if it doesn't.
        """
        try:
            return self._cached_matrices[self._membrane_potential]
        except KeyError:
            matrices = self._model.matrices(
                self._membrane_potential, self._parameters)
            self._cached_matrices[self._membrane_potential] = matrices
            return matrices

    def membrane_potential(self):
        """
        Returns the currently set membrane potential.
        """
        return self._membrane_potential

    def parameters(self):
        """
        Returns the currently set parameter values.
        """
        return list(self._parameters)

    def pre(self, duration):
        """
        Performs an unlogged simulation for ``duration`` time units and uses
        the final state as the new default state.

        After the simulation:

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        Calls to :meth:`reset` after using :meth:`pre` will set the current
        state to this new default state.
        """
        # Check arguments
        duration = float(duration)
        if duration < 0:
            raise ValueError('Duration must be non-negative.')

        # Run
        # This isn't much faster, but this way the simulation's interface is
        # similar to the standard simulation one.
        old_time = self._time
        self.run(duration, log_interval=2 * duration)

        # Update default state
        self._default_state = np.array(self._state, copy=True)

        # Reset time, reset protocol
        self._time = old_time
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to zero.
        - The state is set to the default state.

        """
        self._time = 0
        self._state = np.array(self._default_state, copy=True)
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def run(self, duration, log=None, log_interval=0.01, log_times=None):
        """
        Runs a simulation for ``duration`` time units.

        After the simulation:

        - The simulation time will be increased by ``duration`` time units.
        - The simulation state will be updated to the last reached state.

        Arguments:

        ``duration``
            The number of time units to simulate.
        ``log``
            A log from a previous run can be passed in, in which case the
            results will be appended to this one.
        ``log_interval``
            The time between logged points.
        ``log_times``
            A pre-defined sequence of times to log at. If set, ``log_interval``
            will be ignored.

        Returns a :class:`myokit.DataLog` with the simulation results.
        """
        # Check arguments
        duration = float(duration)
        if duration < 0:
            raise ValueError('Duration must be non-negative.')
        log_interval = float(log_interval)
        if log_interval <= 0:
            raise ValueError('Log interval must be greater than zero.')

        # Check log_times
        if log_times is None:
            log_times = self._time + np.arange(0, duration, log_interval)

        # Set up logging
        if log is None:

            # Create new log
            log = myokit.DataLog()
            log.set_time_key(self._time_key)
            for key in self._log_keys:
                log[key] = np.zeros(log_times.shape)
            offset = 0

        else:

            # Check existing log
            if len(log.keys()) > len(self._log_keys):
                raise ValueError('Invalid log: contains extra keys.')
            try:
                key = self._time_key    # Note: error msg below uses `key`
                offset = len(log[key])
                for key in self._log_keys:
                    log[key] = np.concatenate((
                        log[key], np.zeros(log_times.shape)))
            except KeyError:
                raise ValueError(
                    'Invalid log: missing entry for <' + str(key) + '>.')

        # Run simulation
        if self._protocol is None:

            # User defined membrane potential
            self._run(log, log_times, self._time + duration, offset)

        else:

            # Voltage clamp
            tfinal = self._time + duration
            while self._time < tfinal:
                # Run simulation
                tnext = min(tfinal, self._pacing.next_time())
                times = log_times[np.logical_and(
                    log_times >= self._time, log_times < tnext)]
                self._run(log, times, tnext, offset)
                offset += len(times)

                # Update pacing
                self._membrane_potential = self._pacing.advance(tnext)

        # Return
        return log

    def _run(self, log, times, tnext, offset):
        """
        Runs a simulation with the current membrane potential.

        Arguments:

        ``log``
            The log to append to.
        ``times``
            The times to evaluate at.
        ``tnext``
            The final time to move to.
        ``offset``
            The offset in the ``times`` array to log in

        """
        # Simulate with fixed V
        if self._has_current:
            states, currents = self.solve(times - self._time)
        else:
            states = self.solve(times - self._time)

        # Log results
        lo = offset
        hi = lo + len(times)
        key = log.time_key()
        log[key][lo:hi] = times
        for i, key in enumerate(self._model.states()):
            log[key][lo:hi] = states[i]
        if self._has_current:
            key = self._model.current()
            log[key][lo:hi] = currents
        key = self._model._membrane_potential
        log[key][lo:hi] = self._membrane_potential

        # Now run simulation for final time (which might not be included in the
        # list of logged times, and should not, if you want to be able to
        # append logs without creating duplicate points).
        times = np.array([tnext - self._time])
        if self._has_current:
            states, currents = self.solve(times)
        else:
            states = self.solve(times)

        # Update simulation state
        self._state = np.array(states[:, -1], copy=True)
        self._time = tnext

    def set_constant(self, variable, value):
        """
        Updates a single parameter to a new value.
        """
        self._parameters[self._parameter_map[variable]] = float(value)

    def set_default_state(self, state):
        """
        Changes this simulation's default state.
        """
        state = np.array(state, copy=True, dtype=float)
        if len(state) != len(self._state):
            raise ValueError(
                'Wrong size state vector, expecing (' + str(len(self._state))
                + ') values.')
        if np.any(state < 0):
            raise ValueError(
                'The fraction of channels in a state cannot be negative.')
        if np.abs(np.sum(state) - 1) > 1e-6:
            raise ValueError('The values in `state` must sum to 1.')
        self._default_state = state

    def set_membrane_potential(self, v):
        """
        Changes the membrane potential used in this simulation.
        """
        if self._protocol:
            raise Exception(
                'Membrane potential cannot be set if a protocol is used.')
        self._membrane_potential = float(v)

    def set_parameters(self, parameters):
        """
        Changes the parameter values used in this simulation.
        """
        if len(parameters) != len(self._parameters):
            raise ValueError(
                'Wrong size parameter vector, expecting ('
                + str(len(self._parameters)) + ') values.')
        self._parameters = np.array(parameters, copy=True, dtype=float)

        # Invalidate cache
        self._cached_matrices = {}
        self._cached_solution = {}

    def set_state(self, state):
        """
        Changes the initial state used by in this simulation.
        """
        state = np.array(state, copy=True, dtype=float)
        if len(state) != len(self._state):
            raise ValueError(
                'Wrong size state vector, expecing (' + str(len(self._state))
                + ') values.')
        if np.any(state < 0):
            raise ValueError(
                'The fraction of channels in a state cannot be negative.')
        if np.abs(np.sum(state) - 1) > 1e-6:
            raise ValueError('The values in `state` must sum to 1.')
        self._state = state

    def solve(self, times):
        """
        Evaluates and returns the states at the given times.

        In contrast to :meth:`run()`, this method simply evaluates the states
        (and current) at the given times, using the last known settings for
        the state and membrane potential. It does not use a protocol and does
        not take into account the simulation time. After running this method,
        the state and simulation time are *not* updated.

        Arguments:

        ``times``
            A series of times, where each time must be some ``t >= 0``.

        For models with a current variable, this method returns a tuple
        ``(state, current)`` where ``state`` is a matrix of shape
        ``(len(states), len(times))`` and ``current`` is a vector
        of length ``len(times)``.

        For models without a current variable, only ``state`` is returned.
        """
        n = len(self._state)

        # Solve system, or get cached solution
        try:
            E, P, PI, B = self._cached_solution[self._membrane_potential]
        except KeyError:
            # Get matrices
            A, B = self._matrices()

            # Get eigenvalues, matrix of eigenvectors
            E, P = np.linalg.eig(A)
            E = E.reshape((n, 1))
            PI = np.linalg.inv(P)

            # Cache results
            self._cached_solution[self._membrane_potential] = (E, P, PI, B)

        # Calculate transform of initial state
        y0 = PI.dot(self._state.reshape((n, 1)))

        # Reshape times array
        times = np.array(times, copy=False).reshape((len(times),))

        # Calculate state
        x = P.dot(y0 * np.exp(times * E))

        # Calculate current and/or return
        if self._has_current:
            return x, B.dot(x)
        else:
            return x

    def state(self):
        """
        Returns the initial state used by this simulation.
        """
        return list(self._state)


class DiscreteSimulation:
    """
    Performs stochastic simulations of a :class:`LinearModel`'s behavior for a
    finite number of channels.

    Simulations are run using the "Direct method" proposed by Gillespie [1].

    Each simulation object maintains an internal state consisting of

    * The current simulation time
    * The current state
    * The default state

    When a simulation is created, the simulation time is set to zero and both
    the current and default state are initialized using the ``LinearModel``.
    After each call to :meth:`run()` the time and current state are updated,
    so that each successive call to run continues where the previous simulation
    left off.

    A :class:`protocol <myokit.Protocol>` can be used to set the membrane
    potential during the simulation, or the membrane potential can be adjusted
    manually between runs.

    Example::

        import myokit
        import myokit.lib.markov as markov

        # Create linear markov model
        m = myokit.load_model('clancy-1999.mmt')
        m = markov.LinearModel.from_component(m.get('ina'))

        # Run discrete simulation
        s = markov.DiscreteSimulation(m, nchannels=1000)
        s.set_membrane_potential(-30)
        d = s.run(10)

        import matplotlib.pyplot as plt
        plt.figure()
        for state in m.states():
            plt.step(d.time(), d[state], label=state)
        plt.legend()
        plt.show()

    References

    [1] Gillespie (1976) A General Method for Numerically Simulating the
        stochastic time evolution of coupled chemical reactions
        The Journal of Computational Physics, 22, 403-434.

    Arguments:

    ``model``
        A :class:`LinearModel`.
    ``nchannels``
        The number of channels to simulate.

    """

    def __init__(self, model, protocol=None, nchannels=100):
        # Check model
        if not isinstance(model, LinearModel):
            raise ValueError('First parameter must be a `LinearModel`.')
        self._model = model

        # Check protocol
        if protocol is None:
            self._protocol = None
        elif not isinstance(protocol, myokit.Protocol):
            raise ValueError('Protocol must be a myokit.Protocol object')
        else:
            self._protocol = protocol.clone()

        # Get state and discretize
        nchannels = int(nchannels)
        if nchannels < 1:
            raise ValueError('The number of channels must be at least 1.')
        self._nchannels = nchannels

        # Set state
        self._state = self.discretize_state(self._model.default_state())

        # Set default state
        self._default_state = list(self._state)

        # Set membrane potential
        self._membrane_potential = self._model.default_membrane_potential()

        # Set parameters
        self._parameters = np.array(
            self._model.default_parameters(), copy=True, dtype=float)

        # Mapping from parameter names to index in parameter array
        self._parameter_map = {}
        for i, p in enumerate(self._model.parameters()):
            self._parameter_map[p] = i

        # Cached transition rate list & current matrix
        self._cached_rates = None
        self._cached_matrix = None

        # Set simulation time
        self._time = 0

        # If protocol was given, create pacing system, update vm
        self._pacing = None
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def default_state(self):
        """
        Returns the default simulation state.
        """
        return list(self._default_state)

    def discretize_state(self, x):
        """
        Converts a list of fractional state occupancies to a list of channel
        counts.

        Arguments:

        ``x``
            A fractional state where ``sum(x) == 1``.

        Returns a discretized state ``y`` where ``sum(y) = nchannels``.
        """
        x = np.array(x, copy=False, dtype=float)
        if (np.abs(1 - np.sum(x))) > 1e-6:
            raise ValueError(
                'The sum of fractions in the state to be discretized must'
                ' equal 1.')
        y = np.round(x * self._nchannels)
        # To make sure it always sums to 1, correct the value found at the
        # index with the biggest rounding error.
        i = np.argmax(np.abs(x - y))
        y[i] = 0
        y[i] = self._nchannels - np.sum(y)
        return list(y)

    def membrane_potential(self):
        """
        Returns the current membrane potential.
        """
        return self._membrane_potential

    def number_of_channels(self):
        """
        Returns the number of channels used in this simulation.
        """
        return self._nchannels

    def parameters(self):
        """
        Returns the current parameter values.
        """
        return list(self._parameters)

    def pre(self, duration):
        """
        Performs an unlogged simulation for ``duration`` time units and uses
        the final state as the new default state.

        After the simulation:

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.

        Calls to :meth:`reset` after using :meth:`pre` will set the current
        state to this new default state.
        """
        # Check arguments
        duration = float(duration)
        if duration < 0:
            raise ValueError('Duration must be non-negative.')

        # Run
        # This isn't much faster, but this way the simulation's interface is
        # similar to the standard simulation one.
        old_time = self._time
        self.run(duration)

        # Update default state
        self._default_state = list(self._state)

        # Reset time, reset protocol
        self._time = old_time
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def _current_matrix(self):
        """
        Returns the (cached or regenerated) current matrix B.
        """
        if self._cached_matrix is None:
            self._cached_matrix = self._model.matrices(
                self._membrane_potential, self._parameters)[1]
        return self._cached_matrix

    def _rates(self):
        """
        Returns the (cached or regenerated) transition rate list.
        """
        if self._cached_rates is None:
            self._cached_rates = self._model.rates(
                self._membrane_potential, self._parameters)
        return self._cached_rates

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to zero.
        - The state is set to the default state.

        """
        self._time = 0
        self._state = list(self._default_state)
        if self._protocol:
            self._pacing = myokit.PacingSystem(self._protocol)
            self._membrane_potential = self._pacing.advance(self._time)

    def run(self, duration, log=None):
        """
        Runs a simulation for ``duration`` time units.

        After the simulation:

        - The simulation time will be increased by ``duration`` time units.
        - The simulation state will be updated to the last reached state.

        Arguments:

        ``duration``
            The number of time units to simulate.
        ``log``
            A log from a previous run can be passed in, in which case the
            results will be appended to this one.

        Returns a :class:`myokit.DataLog` with the simulation results.
        """
        # Check arguments
        duration = float(duration)
        if duration < 0:
            raise ValueError('Duration must be non-negative.')

        # Set up logging
        time_key = self._model._model.time().qname()
        log_vars = [time_key, self._model._membrane_potential]
        log_vars.extend(self._model.states())
        cur_key = self._model.current()
        if cur_key is not None:
            log_vars.append(cur_key)

        if log is None:
            # Create new log
            log = myokit.DataLog()
            log.set_time_key(time_key)
            for var in log_vars:
                log[var] = []

        else:

            # Check existing log
            if len(log.keys()) > len(log_vars):
                raise ValueError('Invalid log: contains extra keys.')
            try:
                for key in log_vars:
                    log[key]
            except KeyError:
                raise ValueError(
                    'Invalid log: missing entry for <' + str(key) + '>.')

        if self._protocol is None:
            # Simulate with fixed V
            self._run(duration, log)

        else:

            # Voltage clamp
            tfinal = self._time + duration
            while self._time < tfinal:
                # Run simulation
                tnext = min(tfinal, self._pacing.next_time())
                self._run(tnext - self._time, log)
                # Update pacing
                self._membrane_potential = self._pacing.advance(tnext)
                self._cached_rates = None
                self._cached_matrix = None

        # Return
        return log

    def _run(self, duration, log):
        """
        Runs a simulation with the current membrane potential.
        """
        # Get logging lists
        log_time = log.time()
        log_states = []
        for key in self._model.states():
            log_states.append(log[key])

        # Get current, time and state
        t = self._time
        state = np.array(self._state, copy=True, dtype=int)

        # Get list of transitions
        R = []      # Transition rates
        SI = []     # From state
        SJ = []     # To state
        for i, j, rij in self._rates():
            SI.append(i)
            SJ.append(j)
            R.append(rij)
        R = np.array(R)
        SI = np.array(SI)
        SJ = np.array(SJ)

        # Run
        n_steps = 0
        t_stop = self._time + duration

        # Request for a short time can result in duration=0 at this point,
        # Must set variables otherwise set in loop below here
        if t >= t_stop:
            lambdas = R * state[SI]

        while t < t_stop:
            # Log
            log_time.append(t)
            for i, x in enumerate(state):
                log_states[i].append(x)
            n_steps += 1

            # Get lambdas
            lambdas = R * state[SI]

            # Get sum of lambdas
            lsum = np.sum(lambdas)

            # Sample time until next transition from an exponential
            # distribution with mean 1 / lsum
            tau = np.random.exponential(1 / lsum)

            # Don't step beyond the stopping time!
            if t + tau > t_stop:
                break

            # Get type of transition
            transition = np.random.uniform(0, lsum)
            rsum = 0
            for i, r in enumerate(lambdas):
                rsum += r
                if rsum > transition:
                    break

            # Perform transition
            state[SI[i]] -= 1
            state[SJ[i]] += 1
            t += tau

        # Perform final step using the "brute-force" approach, ensuring we
        # reach self._time + duration exactly.
        # Note that for large tau, the estimates of the probability that
        # something changes may become inaccurate (and > 1)
        # I didn't see this in testing...
        tau = (self._time + duration) - t
        lambdas *= tau
        for i, r in enumerate(lambdas):
            if np.random.uniform(0, 1) < r:
                # Perform transition
                state[SI[i]] -= 1
                state[SJ[i]] += 1

        # Add vm to log
        vm_key = self._model._membrane_potential
        log[vm_key].extend([self._membrane_potential] * n_steps)

        # Add current to log
        c = self._model.current()
        if c is not None:
            B = self._current_matrix()
            x = np.vstack([
                np.array(log_states[i][-n_steps:], dtype=float)
                for i in range(len(state))])
            x /= self._nchannels
            log[c].extend(B.dot(x))

        # Update current state and time
        self._state = list(state)
        self._time += duration

    def set_constant(self, variable, value):
        """
        Updates a single parameter to a new value.
        """
        self._parameters[self._parameter_map[variable]] = float(value)

    def set_default_state(self, state):
        """
        Changes the default state used in the simulation.
        """
        state = np.asarray(state, dtype=int)
        if len(state) != len(self._state):
            raise ValueError(
                'Wrong size state vector, expecing (' + str(len(self._state))
                + ') values.')
        if np.min(state) < 0:
            raise ValueError(
                'The number of channels in a markov model state can not be'
                ' negative.')
        if np.sum(state) != self._nchannels:
            raise ValueError(
                'The number of channels in the default state vector must'
                ' equal ' + str(self._nchannels) + '.')
        self._default_state = list(state)

    def set_membrane_potential(self, v):
        """
        Changes the membrane potential used in this simulation.
        """
        if self._protocol:
            raise Exception(
                'Membrane potential cannot be set if a protocol is used.')
        self._membrane_potential = float(v)
        self._cached_rates = None
        self._cached_matrix = None

    def set_parameters(self, parameters):
        """
        Changes the parameter values used in this simulation.
        """
        if len(parameters) != len(self._parameters):
            raise ValueError(
                'Wrong size parameter vector, expecting ('
                + str(len(self._parameters)) + ') values.')
        self._parameters = np.array(parameters, copy=True, dtype=float)
        self._cached_rates = None
        self._cached_matrix = None

    def set_state(self, state):
        """
        Changes the current state used in the simulation (i.e. the number of
        channels in every markov model state).
        """
        state = np.asarray(state, dtype=int)
        if len(state) != len(self._state):
            raise ValueError(
                'Wrong size state vector, expecing (' + str(len(self._state))
                + ') values.')
        if np.min(state) < 0:
            raise ValueError(
                'The state must be given as a list of non-negative integers.')
        if np.sum(state) != self._nchannels:
            raise ValueError(
                'The number of channels in the state vector must equal '
                + str(self._nchannels) + '.')
        self._state = list(state)

    def state(self):
        """
        Returns the current simulation state.
        """
        return list(self._state)


class MarkovModel:
    """
    **Deprecated**: This class has been replaced by the classes
    :class:`LinearModel` and :class:`AnalyticalSimulation`. Please update your
    code to use these classes instead. This class will be removed in
    future versions of Myokit.
    """

    @staticmethod
    def from_component(
            component, states=None, parameters=None, current=None, vm=None):
        """
        Creates and returns an :class:`AnalyticalSimulation` using a
        :class:`LinearModel` based on a Myokit model component.
        """
        # Deprecated since 2016-01-25
        import warnings
        warnings.warn(
            'The method `MarkovModel.from_component` is deprecated.'
            ' Please use `LinearModel.from_component` instead.')
        return AnalyticalSimulation(LinearModel.from_component(
            component, states, parameters, current, vm))

    def __new__(self, model, states, parameters=None, current=None, vm=None):
        # Deprecated since 2016-01-25
        import warnings
        warnings.warn(
            'The `MarkovModel` class is deprecated.'
            ' Please use the `LinearModel` class instead.')
        return AnalyticalSimulation(LinearModel(
            model, states, parameters, current, vm))


class LinearModelError(myokit.MyokitError):
    """
    Raised for issues with constructing or using a :class:`LinearModel`.
    """


def _linear_combination(expression, variables):
    """
    Checks if ``expression`` is a linear combination of the given ``variables``
    and returns the multiplier for each variable.

    See :meth:`_linear_combination_terms` for details.
    """
    return _linear_combination_terms(_split_terms(expression), variables)


def _linear_combination_terms(terms, variables):
    """
    Checks if a list of terms forms linear combination of the given
    ``variables`` and returns the multiplier for each variable.

    If ``expression`` is a linear combination ``a1*v1 + a2*v2 + ...`` where
    ``vi`` is a variable in ``variables``, this method returns a list of
    expressions ``[a1, a2, ...]``. The expressions ``ai`` are not guaranteed to
    be constants, but won't contain any references to the variables in
    ``variables``.

    If the expression cannot be written as a linear combination a
    ``ValueError`` is raised.
    """
    # Multiplier for each variable
    multipliers = [None] * len(variables)

    # Check each term multiplies one of the variables
    for term in terms:
        # Split into name and multiplier
        name, multiplier = _split_factor(term, variables)

        # Update the variable's multiplier
        var = name.var()
        i = variables.index(var)
        if multipliers[i] is None:
            multipliers[i] = multiplier
        else:
            multipliers[i] = myokit.Plus(multipliers[i], multiplier)

    # Return obtained multipliers
    return multipliers


def _split_factor(term, variables):
    """
    Splits ``term`` into two parts that can be multiplied together, so that one
    part is a reference to a variable in ``variables``, and the other part has
    no references to variables in ``variables``.

    Returns a tuple ``(name, multiplier)`` where ``name`` is a
    :class:`myokit.Name` referencing a variable in the list ``variables``, and
    where ``multiplier`` is some other expression such that ``name*multiplier``
    is equivalent to the original term.

    If the term can't be split this way, a ``ValueError`` is raised.
    """

    # Check that the term references exactly one variable from ``variables``
    names = set([myokit.Name(var) for var in variables])
    refs = set(term.references())
    n_refs = len(names & refs)
    if n_refs != 1:
        raise ValueError(
            'The expression `term` must reference exactly one variable from'
            ' the list `variables` (found ' + str(n_refs) + ').')

    # Get Name of variable referenced in this term
    name = (names & refs).pop()

    # Split
    m = None
    positive = True
    while term != name:
        t = type(term)
        if t == myokit.PrefixPlus:
            term = term[0]
        elif t == myokit.PrefixMinus:
            positive = not positive
            term = term[0]
        elif t == myokit.Multiply:
            a, b = term
            if name in b.references():
                a, b = b, a
            term = a
            m = b if m is None else myokit.Multiply(m, b)
        elif t == myokit.Divide:
            a, b = term
            if name in b.references():
                raise ValueError(
                    'Non-linear function (division) of ' + str(name)
                    + ' found in ' + str(term) + '.')
            term = a
            m = myokit.Divide(myokit.Number(1) if m is None else m, b)
        elif t in (myokit.Plus, myokit.Minus):
            raise ValueError(
                'Expression passed to _split_factor must be a single term.')
        else:
            raise ValueError(
                'Non-linear function of ' + str(name) + ' found in '
                + str(term) + '.')

    # Finalize multiplier
    if m is None:
        m = myokit.Number(1)
    if not positive:
        m = myokit.PrefixMinus(m)

    # Return
    return name, m


def _split_terms(expression, terms=None, positive=True):
    """
    Takes an expression tree of :class:`myokit.Plus` and/or
    :class:`myokit.Minus` objects and splits it into terms.

    Arguments:

    ``expression``
        The expression to split.
    ``terms``
        Used internally: A list of terms to append to.
    ``positive``
        Used internally: If false, the terms will be multiplied by -1
        before adding.

    """
    if terms is None:
        terms = []
    if type(expression) == myokit.Plus:
        a, b = expression
        _split_terms(a, terms, positive)
        _split_terms(b, terms, positive)
    elif type(expression) == myokit.Minus:
        a, b = expression
        _split_terms(a, terms, positive)
        _split_terms(b, terms, not positive)
    elif type(expression) == myokit.PrefixPlus:
        _split_terms(expression[0], terms, positive)
    elif type(expression) == myokit.PrefixMinus:
        _split_terms(expression[0], terms, not positive)
    else:
        if positive:
            terms.append(expression)
        else:
            terms.append(myokit.PrefixMinus(expression))
    return terms


def find_markov_models(model):
    """
    Searches a :class:`myokit.Model` for groups of states that constitute a
    Markov model.

    Returns a list of lists, where the inner lists are groups of variables that
    form a Markov model together.

    Note that this method performs a shallow check of the equation shapes,
    and does not perform any simplification or rewriting to see if the
    expressions can be made to fit a Markov form.

    Arguments:

    ``model``
        The :class:`myokit.Model` to search.

    """

    # Models
    models = []

    # Scan model for clusters of states that depend on each other
    seen = set()
    for var in model.states():
        if var in seen:
            continue

        # Get references to other states, made by this state
        group = set(var.refs_to(True))

        # Find group of connected states
        todo = collections.deque(group)
        while todo:
            var = todo.popleft()
            for ref in set(var.refs_to(True)) - group:
                group.add(ref)
                todo.append(ref)

        # All these states now count as 'seen'
        seen |= group

        # Now check if there's a (1 - x1 - x2 - ...) variable:
        # First check if there's a non-state variable that depends on all of
        # these states.
        candidates = set()
        for var in group:
            for ref in var.refs_by(True):
                if set(ref.refs_to(True)) == group:
                    candidates.add(ref)
        candidates -= group

        # Now test the candidates for the correct form
        extra = set()
        for candidate in candidates:
            # Split into terms
            terms = _split_terms(candidate.rhs())

            # Find and remove the '1' term
            i_one = None
            for i, term in enumerate(terms):
                if term.is_constant() and term.eval() == 1:
                    i_one = i
                    break
            if i_one is None:
                continue
            del terms[i_one]

            # Remaining terms must be linear combination of the states in
            # group...
            try:
                factors = _linear_combination_terms(terms, list(group))
            except ValueError:
                continue

            # And each factor must be -1
            ok = True
            for factor in factors:
                if not (factor.is_constant() and factor.eval() == -1):
                    ok = False
                    break
            if not ok:
                continue

            # Passed all tests!
            extra.add(candidate)
        del candidates

        # At this point `extra` should be empty or a single variable, if not,
        # it's not a (normal) Markov model
        if len(extra) > 1:
            continue

        # Must have at least 2 states in total
        states = group | extra
        if len(states) < 2:
            continue

        # Get sorted list of states for output
        states = list(states)
        states.sort(key=lambda x: myokit.tools.natural_sort_key(x.qname()))

        # Check all members of `group` are a linear combination of states
        try:
            for state in group:
                _linear_combination(state.rhs(), states)
        except ValueError:
            continue

        # This looks like a Markov model!
        models.append(states)

    return models


def convert_markov_models_to_compact_form(model):
    """
    Scans a :class:`myokit.Model` for Markov models, and ensures they contain
    one state that's not evaluated as an ODE, but as ``1 - sum(x[i])``, where
    the sum is over all other states ``x[i]``.

    Arguments:

    ``model``
        The :class:`myokit.Model` to scan.

    Returns an updated :class:`myokit.Model`.
    """
    # Clone model
    model = model.clone()

    # Find markov models and convert
    for states in find_markov_models(model):

        # Check if a non-ODE state is already present
        if sum([1 for x in states if not x.is_state()]):
            continue

        # Update final state
        state = states[-1]
        state.demote()
        state.set_rhs('1 - ' + '-'.join([x.qname() for x in states[:-1]]))

    return model


def convert_markov_models_to_full_ode_form(model):
    """
    Scans a :class:`myokit.Model` for Markov models, and ensures they are
    written in a form where every Markov state is evaluated as an ODE.

    Arguments:

    ``model``
        The :class:`myokit.Model` to scan.

    Returns an updated :class:`myokit.Model`.
    """
    # Clone model
    model = model.clone()

    # Find markov models and convert
    for states in find_markov_models(model):

        # Find 1-sum() state
        special = None
        i_special = None
        for i, state in enumerate(states):
            if not state.is_state():
                special = state
                i_special = i
                break

        # No special state: then no need to convert
        if special is None:
            continue

        # Get initial value for special state
        initial_value = special.eval()

        # Gather terms for existing states, see which ones don't cancel out
        sum_of_terms = [[] for _ in states]
        for state in states:
            if state is special:
                continue

            factors = _linear_combination(state.rhs(), states)
            for i, factor in enumerate(factors):
                if factor is None:
                    continue

                # Split terms in factor, add each to sum_of_terms
                for term in _split_terms(factor):

                    # Get negative term
                    if isinstance(term, myokit.PrefixMinus):
                        negative = term[0]
                    else:
                        negative = myokit.PrefixMinus(term)

                    # If negative term is in list, then these cancel out and
                    # should be removed
                    try:
                        sum_of_terms[i].remove(negative)
                    except ValueError:
                        # Negative term isn't in the list, so add this factor
                        # to the list.
                        sum_of_terms[i].append(term)

        # Create RHS from remaining terms
        terms = []
        for i, sums in enumerate(sum_of_terms):
            if not sums:
                continue

            # Combine terms
            term = sums[0]
            for t in sums[1:]:
                term = myokit.Plus(term, t)

            # Negate
            if isinstance(term, myokit.PrefixMinus):
                term = term[0]
            else:
                term = myokit.PrefixMinus(term)

            # Multiply with state and store
            terms.append(myokit.Multiply(term, myokit.Name(states[i])))

        # Combine terms (should always be 2 or more).
        rhs = terms[0]
        for term in terms[1:]:
            rhs = myokit.Plus(rhs, term)

        # Update special state
        special.promote(initial_value)
        special.set_rhs(rhs)

    # Return cloned & updated model
    return model

